import os
import re
import gc
import time
import socket
import pprint
import inspect
import threading
import functools
from contextlib import contextmanager

import psutil

try:
    from greenlet import greenlet
except ImportError:
    greenlet = None
try:
    import gevent
    from gevent.socket import socket as gevent_socket
except ImportError:
    gevent_socket = socket.socket
try:
    from gevent.monkey import is_anything_patched
except ImportError:

    def is_anything_patched():
        return False


def _dictobj_replace(obj):
    if isinstance(obj, (str, bytes)):
        s = str(obj[:10])
        return s + "..."
    elif isinstance(obj, bytearray):
        s = str(obj[:10])
        return f"{(s[:-2])}...{(s[-2:])}"
    else:
        return obj


def print_obj(obj, indent=0):
    tab = " " * indent
    if inspect.ismodule(obj):
        print(f"{tab}+ {repr(obj)}")
    else:
        print(f"{tab}+ <{type(obj).__name__} object at 0x{'{:x}'.format(id(obj))}>")
        if isinstance(obj, dict):
            obj = {k: _dictobj_replace(v) for k, v in obj.items()}
        s = "".join(pprint.pformat(obj, indent=indent + 2, compact=True, depth=1))
        print(f"{tab}|_{s}")


def printobj_backref(queue, max_depth=10):
    ignore = set()
    ignore.add(id(ignore))
    ignore.add(id(queue))
    queue = [(obj, 0) for obj in queue]
    ignore.add(id(queue))
    for tpl in queue:
        ignore.add(id(tpl))
    gc.collect()
    while queue:
        target, depth = queue.pop()
        ignore.add(id(target))
        if depth > max_depth:
            continue
        print_obj(target, indent=depth)
        if inspect.ismodule(target):
            continue

        refs = gc.get_referrers(target)
        ignore.add(id(refs))
        for ref in refs:
            if id(ref) in ignore:
                continue
            tpl = (ref, depth + 2)
            ignore.add(id(tpl))
            queue.append(tpl)


def gc_collect(*classes):
    """
    :param classes:
    :returns set:
    """
    ret = []
    if not classes:
        return ret
    # TODO: gevent monitoring says this blocks
    # find a better way?
    yield_to_eventloop = is_anything_patched()
    for ob in gc.get_objects():
        if yield_to_eventloop:
            gevent.sleep()
        try:
            if not isinstance(ob, classes):
                continue
        except ReferenceError:
            continue
        ret.append(ob)
    return ret


def file_descriptors(pid=None):
    """
    All file descriptors of a process

    :param int pid: current by default
    :returns dict:
    """
    if pid is None:
        pid = os.getpid()
    fdpath = os.path.join(psutil.PROCFS_PATH, str(pid), "fd")
    fds = {}
    for fd in os.listdir(fdpath):
        try:
            dest = os.readlink(os.path.join(fdpath, fd))
        except Exception:
            pass
        else:
            fds[int(fd)] = dest
    return fds


def greenlets():
    """
    All greenlets of this process

    :returns set:
    """
    if greenlet is None:
        return set()
    return set(gc_collect(greenlet))


def sockets():
    """
    All sockets of this process

    :returns set:
    """
    return set(gc_collect(gevent.socket.socket, socket.socket))


def resources(*classes):
    """
    All other resource of this process

    :returns set:
    """
    return set(gc_collect(*classes))


def threads(pid=None):
    """
    All threads of a process

    :param int pid: current by default
    :returns set:
    """
    if pid is None or pid == os.getpid():
        return set(threading.enumerate())
    else:
        process = psutil.Process(pid=pid)
        return {t.id for t in process.threads()}


def memory(pid=None):
    """
    Used memory of a process

    :param int pid: current by default
    :returns int: bytes
    """
    process = psutil.Process(pid=pid)
    return process.memory_info().rss


def file_processes(pattern):
    r"""
    All processes that have a file opened

    Example: all open hdf5 files

    .. code-block:: python
        file_processes(r".+\.h5$")

    :param str pattern: regex pattern
    :returns list((str, Process)):
    """
    ret = []
    for proc in psutil.process_iter():
        try:
            fds = file_descriptors(proc.pid)
            for filename in fds.values():
                if re.match(pattern, filename):
                    ret.append((filename, proc))
        except Exception:
            pass
    return ret


def log_file_processes(logfunc, pattern):
    """
    All processes that have a file opened

    :param callable logfunc:
    :param str pattern: regex pattern
    """
    procs = file_processes(pattern)
    if not procs:
        return
    pid = os.getpid()
    msg = "\n### Files matching {} opened by {} process".format(
        repr(pattern), len(procs)
    )
    for fname, proc in procs:
        cmdline = repr(" ".join(proc.cmdline()))
        if proc.pid == pid:
            proc = "CURRENT {}".format(proc)
        msg += "\nFile {}:\n Opened by {}\n {}".format(repr(fname), proc, cmdline)
    msg += "\n###\n"
    logfunc(msg)


def matching_processes(pattern):
    """
    Find processes

    :param str pattern: regex pattern
    :yields Process:
    """
    for proc in psutil.process_iter():
        if re.match(pattern, " ".join(proc.cmdline())):
            yield proc


def terminate(pattern, wait=False):
    """
    Terminate a process

    :param str pattern: regex pattern
    :param bool force:
    """
    for proc in matching_processes(pattern):
        proc.terminate()
        if wait:
            proc.wait()


def kill(pattern, wait=False):
    """
    Terminate a process

    :param str pattern: regex pattern
    :param bool force:
    """
    for proc in matching_processes(pattern):
        proc.kill()
        if wait:
            proc.wait()


def log_diff(logfunc, difffunc, typ, diffargs=None, prefix=""):
    """
    :param callable logfunc:
    :param callable difffunc:
    :param str typ:
    :param tuple diffargs:
    :param str prefix:
    """
    new, diff = difffunc(*diffargs)
    if diff:
        if prefix:
            prefix = prefix.format(len(diff)) + ": "
        else:
            prefix = f"{len(diff)} {typ} difference "
        try:
            lst = sorted(diff)
        except TypeError:
            lst = list(diff)
        printobj_backref(lst)
        lst = list(map(str, lst))
        logfunc(f"{prefix}{lst}")
    return new


def _fd_diff(old):
    """New file descriptors

    :param dict old:
    :returns dict, set:
    """
    new = file_descriptors()
    diff = set(new.values()) - set(old.values())
    return new, diff


def _class_diff(old, classes):
    """New instance diff

    :param set old:
    :param tuple classes:
    :returns set, set:
    """
    new = set(gc_collect(*classes))
    return new, new - old


def _greenlet_diff(old):
    """
    New greenlets

    :param set old:
    :returns set, set:
    """
    new = greenlets()
    return new, new - old


def _socket_diff(old):
    """
    New sockets

    :param set old:
    :returns set, set:
    """
    new = sockets()
    return new, new - old


def _thread_diff(old):
    """
    New threads

    :param set old:
    :returns set, set:
    """
    new = threads()
    return new, new - old


def log_fd_diff(logfunc, old, prefix=""):
    """
    Print new file descriptors

    :param callable logfunc:
    :param dict old:
    :param str prefix:
    :returns dict:
    """
    return log_diff(logfunc, _fd_diff, "fds", diffargs=(old,), prefix=prefix)


def log_greenlet_diff(logfunc, old, prefix=""):
    """
    Print new greenlets

    :param callable logfunc:
    :param set old:
    :param str prefix:
    :returns set:
    """
    return log_diff(
        logfunc, _greenlet_diff, "greenlets", diffargs=(old,), prefix=prefix
    )


def log_class_diff(logfunc, old, classes, prefix=""):
    """Print new class instances

    :param callable logfunc:
    :param set old:
    :param str prefix:
    :returns set:
    """
    return log_diff(
        logfunc, _class_diff, str(classes), diffargs=(old, classes), prefix=prefix
    )


def log_socket_diff(logfunc, old, prefix=""):
    """
    Print new sockets

    :param callable logfunc:
    :param set old:
    :param str prefix:
    :returns set:
    """
    return log_diff(logfunc, _socket_diff, "sockets", diffargs=(old,), prefix=prefix)


def log_thread_diff(logfunc, old, prefix=""):
    """
    Print new threads

    :param callable logfunc:
    :param set old:
    :param str prefix:
    :returns set:
    """
    return log_diff(logfunc, _thread_diff, "threads", diffargs=(old,), prefix=prefix)


def _diff_exception(*args, **kwargs):
    raise AssertionError(*args, **kwargs)


def assert_fd_diff(old, prefix=""):
    """
    Raise exception when new file descriptors

    :param dict old:
    :param str prefix:
    :returns dict:
    """
    return log_fd_diff(_diff_exception, old, prefix=prefix)


def assert_greenlet_diff(old, prefix=""):
    """
    Raise exception when new greenlets

    :param set old:
    :param str prefix:
    :returns set:
    """
    return log_greenlet_diff(_diff_exception, old, prefix=prefix)


def assert_class_diff(old, classes, prefix=""):
    """
    Raise exception when new instances

    :param set old:
    :param str prefix:
    :returns set:
    """
    return log_class_diff(_diff_exception, old, classes, prefix=prefix)


def assert_socket_diff(old, prefix=""):
    """
    Raise exception when new sockets

    :param set old:
    :param str prefix:
    :returns set:
    """
    return log_socket_diff(_diff_exception, old, prefix=prefix)


def assert_thread_diff(old, prefix=""):
    """
    Raise exception when new threads

    :param set old:
    :param str prefix:
    :returns set:
    """
    return log_thread_diff(_diff_exception, old, prefix=prefix)


class ResourceMonitor:
    def __init__(self, *classes, greenlets=True, threads=True, fds=True, sockets=True):
        self._yield_to_eventloop = is_anything_patched()
        self.classes = classes
        if classes:
            self.others = set()
        else:
            self.others = None
        if greenlets:
            self.greenlets = set()
        else:
            self.greenlets = None
        if threads:
            self.threads = set()
        else:
            self.threads = None
        if fds:
            self.fds = set()
        else:
            self.fds = None
        if sockets:
            self.sockets = set()
        else:
            self.sockets = None

    def start(self):
        if self._yield_to_eventloop:
            gevent.get_hub()
        if self.greenlets is not None:
            self.greenlets = greenlets()
        if self.threads is not None:
            self.threads = threads()
        if self.fds is not None:
            self.fds = file_descriptors()
        if self.sockets is not None:
            self.sockets = sockets()
        if self.others is not None:
            self.others = resources(*self.classes)

    @contextmanager
    def check_leaks_context(self, msg=None):
        self.start()
        try:
            yield
        finally:
            self.check_leaks(msg=msg)

    def wait_gc_collect(self):
        time.sleep(0.1)
        while gc.collect():
            time.sleep(0.1)

    def check_leaks(self, msg=None):
        self.wait_gc_collect()
        if self.others is not None:
            assert_class_diff(self.others, self.classes, prefix=msg)
        if self.sockets is not None:
            assert_socket_diff(self.sockets, prefix=msg)
        if self.fds is not None:
            assert_fd_diff(self.fds, prefix=msg)
        if self.threads is not None:
            assert_thread_diff(self.threads, prefix=msg)
        if self.greenlets is not None:
            assert_greenlet_diff(self.greenlets, prefix=msg)


def check_resource_leaks(*classes, **kw):
    """Decorator for function resource checking"""

    def _check_resource_leaks(method):
        @functools.wraps(method)
        def inner(*args, **kwargs):
            mon = ResourceMonitor(*classes, **kw)
            with mon.check_leaks_context(msg=method.__name__):
                method(*args, **kwargs)

        return inner

    return _check_resource_leaks



import logging
from .items import Hdf5ItemWriter
from ..io import nexus


logger = logging.getLogger(__name__)


class Hdf5ScanReferenceWriter(Hdf5ItemWriter):
    """Manages creation and growth one HDF5 group for scan references."""

    def __init__(
        self,
        filename=None,
        parent=None,
        filecontext=None,
        nreferences=0,
        parent_logger=None,
    ):
        """
        :param str filename: HDF5 file name
        :param str filecontext: HDF5 open context manager
        :param str parent: path in the HDF5 file
        :param int nreferences: variable length by default
        :param parent_logger:
        """
        if parent_logger is None:
            parent_logger = logger
        super().__init__(
            filename=filename,
            parent=parent,
            filecontext=filecontext,
            parent_logger=parent_logger,
        )
        self.nreferences = nreferences

    @property
    def name(self):
        return ""

    @property
    def alias(self):
        return ""

    @property
    def npoints_expected(self):
        return self.nreferences

    def add_references(self, newuris):
        """
        Add uri links

        :param list(str) newuris:
        """
        self.add(newuris)

    def _insert_data(self, group, newuris):
        """Add uri links

        :param list(str) newuris:
        :returns int: added links
        """
        duri = nexus.getUri(group)
        for uri in newuris:
            linkname = nexus.hdf5_basename(uri)
            if uri == nexus.hdf5_join(duri, linkname):
                continue  # ignore self-reference
            nexus.createLink(group, linkname, uri)
        return len(newuris)

    def _create_parent(self, nxroot):
        """Create the group which will contain the links"""
        grp = nxroot.create_group(self.parent)
        grp.attrs["NX_class"] = "NXcollection"
        self._parent_exists = True

    def _create(self, nxroot):
        """Create the group which will contain the links"""
        self._create_parent(nxroot)
        self._exists = True



import os
import errno
import random
import tempfile as _tempfile
import string


def tempname(
    size=6, chars=string.ascii_lowercase + string.digits, prefix="", suffix=""
):
    """
    Random name with prefix and suffix
    """
    # Number of combinations: n^size  (default: 62^6)
    name = "".join(random.choice(chars) for _ in range(size))
    return prefix + name + suffix


def temproot():
    """
    OS tmp directory
    """
    return _tempfile.gettempdir()


def tempdir(root=None, **kwargs):
    """
    Random directory in OS tmp directory
    """
    if not root:
        root = temproot()
    return os.path.join(root, tempname(**kwargs))


tempfile = tempdir


def mkdir(path):
    """
    Create directory recursively and silent when already existing

    :param str path:
    """
    if path:
        path = os.path.abspath(path)
        os.makedirs(path, exist_ok=True)


def close_files(*fds):
    exceptions = []
    for fd in fds:
        try:
            if fd is None:
                continue
            try:
                os.close(fd)
            except OSError as e:
                if e.errno == errno.EBADF:
                    pass
                else:
                    raise
        except BaseException as e:
            exceptions.append(e)
    if exceptions:
        raise Exception(exceptions)


def rotatefiles(filename, nmax=10):
    """
    Rename or delete existing file.

    :param str filename:
    :param int nmax:
    """
    mkdir(os.path.dirname(filename))
    filenamegen = rotatefiles_gen(filename, nmax)
    _rotatefiles(filename, filenamegen)


def _rotatefiles(filename, filenamegen):
    """
    Rename or delete existing file

    :param str filename:
    :param generator filenamegen:
    """
    if os.path.exists(filename):
        try:
            nextname = next(filenamegen)
        except StopIteration:
            os.remove(filename)
        else:
            _rotatefiles(nextname, filenamegen)
            os.rename(filename, nextname)


def rotatefiles_gen(filename, nmax):
    """
    Generate rotating file names (maximal nmax files)

    :param str filename:
    :param int nmax:
    """
    if nmax > 1:
        base, ext = os.path.splitext(filename)
        fmt = "{}.{{}}{}".format(base, ext)
        for i in range(1, nmax):
            yield fmt.format(i)



import io
import builtins
import traceback
from functools import wraps
from . import monkey
from ..utils.logging_utils import print_err


OPEN_FILES = {}


def _patch_close(f):
    old_close = f.close

    @wraps(old_close)
    def close():
        name = f.name
        old_close()
        OPEN_FILES.pop(f)
        print_open_files(" AFTER CLOSING " + repr(name))

    f.close = close


def _tracking_open(old_open):
    @wraps(old_open)
    def tracking_open(*args, **kwargs):
        stack = traceback.extract_stack()
        stack.pop()
        f = old_open(*args, **kwargs)
        _patch_close(f)
        OPEN_FILES[f] = stack
        print_open_files(" AFTER OPENING " + repr(f.name))
        return f

    return tracking_open


def print_open_files(msg="", **kwargs):
    print_err("\n### {} OPEN FILES{}".format(len(OPEN_FILES), msg), **kwargs)
    sep = ""
    for f, stack in OPEN_FILES.items():
        stack = sep.join(traceback.format_list(stack))
        print_err("\n Open file {}:\n{}{}".format(repr(f.name), sep, stack), **kwargs)
    print_err("\n### END OPEN FILES\n", **kwargs)


def patch():
    newitem = _tracking_open(monkey.original(builtins, "open"))
    monkey.patch_item(builtins, "open", newitem)
    newitem = _tracking_open(monkey.original(io, "open"))
    monkey.patch_item(io, "open", newitem)
    # Cannot patch these:
    # monkey.patch_item(h5py.h5f, "open")
    # monkey.patch_item(h5py.h5f, "create")


def unpatch():
    monkey.unpatch_item(builtins, "open")
    monkey.unpatch_item(io, "open")
    # monkey.unpatch_item(h5py.h5f, "open")
    # monkey.unpatch_item(h5py.h5f, "create")



import sys
from blisswriter.parameters import all_cli_saveoptions
from blisswriter.subscribers.session_subscriber import NexusSessionSubscriber
from blisswriter.utils import logging_utils
from .utils import config_root_logger


def main(argv=None):
    import argparse

    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(
        description="Start python a process for saving Bliss data in HDF5"
    )
    parser.add_argument("session_name", type=str, help="Session name")
    cli_saveoptions = all_cli_saveoptions()
    for attr, okwargs in cli_saveoptions.items():
        parser.add_argument("--" + attr, **okwargs)
    logging_utils.add_cli_args(parser)

    # Parse CLI arguments
    args = parser.parse_args(argv[1:])
    config_root_logger(args)
    kwargs = {}
    cli_saveoptions = all_cli_saveoptions(configurable=args.configurable)
    for attr, okwargs in cli_saveoptions.items():
        option = okwargs["dest"]
        try:
            kwargs[option] = getattr(args, option)
        except AttributeError:
            continue

    # Launch the session writer
    subscriber = NexusSessionSubscriber(
        args.session_name, writer_name="external", **kwargs
    )
    subscriber.start()
    subscriber.join()


if __name__ == "__main__":
    sys.exit(main())



import enum
from typing import NamedTuple


@enum.unique
class ChannelDataType(enum.IntEnum):
    SCAN_REFERENCE = enum.auto()
    LIMA_STATUS = enum.auto()
    NUMERIC_DATA = enum.auto()


class Channel(NamedTuple):
    name: str
    data_type: ChannelDataType
    info: dict



import os
import multiprocessing
from contextlib import contextmanager

import pytest
import numpy
import h5py.h5t
from silx.io import dictdump
from silx.io.utils import h5py_read_dataset, h5py_read_attribute

from blisswriter.io import nexus


@contextmanager
def nxroot(path, name):
    filename = os.path.join(str(path), name + ".h5")
    rootattrs = {
        "creator": "test",
        "creator_version": "1.0.0",
        "custom_root_attr": "custom_root_attr",
    }
    with nexus.nxRoot(filename, mode="a", rootattrs=rootattrs) as f:
        yield f


def test_nexus_root(tmpdir):
    with nxroot(tmpdir, "test_nexus_root") as h5group:
        validateNxRoot(h5group)


def test_nexus_entry(tmpdir):
    with nxroot(tmpdir, "test_nexus_entry") as h5group:
        entry = nexus.nxEntry(h5group, "entry0001")
        nexus.updated(entry, final=True)
        with pytest.raises(RuntimeError):
            nexus.nxEntry(entry, "entry0002")
        validateNxEntry(entry)


def test_nexus_entry_order(tmpdir):
    with nxroot(tmpdir, "test_nexus_entry") as h5group:
        for i in range(100):
            nexus.nxEntry(h5group, str(i))
            nexus.updateAttribute(h5group, "default", str(i))
        lst = [int(i) for i in h5group.keys()]
        assert lst == list(range(100))
        if nexus.ISSUE_1641_FIXED:
            assert h5group.attrs["default"] == "99"
        else:
            with pytest.raises(AssertionError):
                assert h5group.attrs["default"] == "99"


def test_nexus_process(tmpdir):
    with nxroot(tmpdir, "test_nexus_process") as h5group:
        entry = nexus.nxEntry(h5group, "entry0001")
        configdict = {"a": 1, "b": 2}
        for i, type in enumerate(["json", "ini", None]):
            process = nexus.nxProcess(
                entry, "process{:04d}".format(i), configdict=configdict, type=type
            )
            with pytest.raises(RuntimeError):
                nexus.nxProcess(h5group, "process0002", configdict=configdict)
            validateNxProcess(process)


def test_nexus_data(tmpdir):
    with nxroot(tmpdir, "test_nexus_data") as root:
        entry = nexus.nxEntry(root, "entry0001")
        process = nexus.nxProcess(entry, "process0001")
        data = nexus.nxData(process["results"], "data")
        s = (4, 3, 2)
        datadict = {
            "Fe K": numpy.arange(numpy.prod(s), dtype=float).reshape(s),
            "Ca K": numpy.arange(numpy.prod(s)).reshape(s) + 1,
            "S K": numpy.zeros(s),
        }
        axes = [
            ("y", numpy.arange(s[0]), {"units": "um"}),
            ("x", numpy.arange(s[1]), {}),
            ("z", {"shape": (s[2],), "dtype": int}, None),
        ]
        signals = [
            ("Fe K", datadict["Fe K"], {"interpretation": "image"}),
            ("Ca K", {"data": datadict["Ca K"]}, {}),
            ("S K", {"shape": s, "dtype": int}, None),
        ]
        nexus.nxDataAddAxes(data, axes)
        nexus.nxDataAddSignals(data, signals)

        validateNxData(data, axes, signals)
        signals = nexus.nxDataGetSignals(data)
        assert signals == ["Fe K", "Ca K", "S K"]

        nexus.markDefault(data["Ca K"])
        default = data.file[nexus.getDefault(data.file, signal=False)]
        default = nexus.dereference(default, current=root)
        assert default == nexus.getUri(data)
        default = data.file[nexus.getDefault(data.file, signal=True)]
        default = nexus.dereference(default, current=root)
        assert default == nexus.getUri(data["Ca K"])

        data = entry[nexus.DEFAULT_PLOT_NAME]
        signals = nexus.nxDataGetSignals(data)
        assert signals == ["Ca K", "Fe K", "S K"]
        assert data["y"].attrs["units"] == "um"
        assert data["Fe K"].attrs["interpretation"] == "image"
        for name in signals:
            assert data[name].shape == s
        for n, name in zip(s, list(next(iter(zip(*axes))))):
            assert data[name].shape == (n,)

        # Test dataset concatenation
        def vdatanamegen():
            c = 0
            while True:
                yield "vdata{}".format(c)
                c += 1

        vdataname = vdatanamegen()
        for virtual in False, True:
            value = {
                "axis": 0,
                "newaxis": True,
                "virtual": virtual,
                "data": [nexus.getUri(data[name]) for name in datadict],
            }
            _tmp = next(vdataname)
            vdata = nexus.nxCreateDataSet(process, _tmp, value, None)
            if virtual:
                assert nexus.vdsIsValid(vdata)
            for i, name in enumerate(datadict):
                numpy.testing.assert_array_equal(datadict[name], vdata[i])
            value["axis"] = 1
            vdata1 = nexus.nxCreateDataSet(process, next(vdataname), value, None)
            if virtual:
                assert nexus.vdsIsValid(vdata1)
            for i, name in enumerate(datadict):
                numpy.testing.assert_array_equal(datadict[name], vdata1[:, i])
            value["axis"] = 0
            value["newaxis"] = False
            vdata = nexus.nxCreateDataSet(process, next(vdataname), value, None)
            if virtual:
                assert nexus.vdsIsValid(vdata)
            for i, name in enumerate(datadict):
                numpy.testing.assert_array_equal(
                    datadict[name], vdata[i * s[0] : (i + 1) * s[0]]
                )
            value["axis"] = 1
            vdata = nexus.nxCreateDataSet(process, next(vdataname), value, None)
            if virtual:
                assert nexus.vdsIsValid(vdata)
            for i, name in enumerate(datadict):
                numpy.testing.assert_array_equal(
                    datadict[name], vdata[:, i * s[1] : (i + 1) * s[1]]
                )
            value["data"].append(nexus.getUri(data["y"]))
            with pytest.raises(RuntimeError):
                nexus.nxCreateDataSet(process, next(vdataname), value, None)


def test_nexus_StringAttribute(tmpdir):
    check_string_types(tmpdir, attribute=True, raiseExtended=True)


def test_nexus_StringDataset(tmpdir):
    check_string_types(tmpdir, attribute=False, raiseExtended=True)


def test_nexus_ExtStringAttribute(tmpdir):
    check_string_types(tmpdir, attribute=True, raiseExtended=False)


def test_nexus_ExtStringDataset(tmpdir):
    check_string_types(tmpdir, attribute=False, raiseExtended=False)


def test_nexus_uri(tmpdir):
    path = str(tmpdir)

    uri = "test1.h5::/a::/b"
    a, b = nexus.splitUri(uri)
    assert a == "test1.h5"
    assert b == "/a::/b"

    uri = nexus.normUri("./test1.h5::/a/../b")
    assert uri == "test1.h5::/b"

    uri = "test1.h5::/a/b"
    uriref = "test1.h5::/a"
    a, b = nexus.relUri(uri, uriref)
    assert a == "."
    assert b == "b"

    uri = os.path.join(path, "test1.h5::/a/b")
    uriref = os.path.join(path, "test1.h5::/a")
    a, b = nexus.relUri(uri, uriref)
    assert a == "."
    assert b == "b"

    uri = "test1.h5::/a/b"
    uriref = "test2.h5::/a"
    a, b = nexus.relUri(uri, uriref)
    assert a == "test1.h5"
    assert b == "/a/b"

    uri = os.path.join(path, "test1.h5::/a/b")
    uriref = os.path.join(path, "test2.h5::/a")
    a, b = nexus.relUri(uri, uriref)
    assert a == "./test1.h5"
    assert b == "/a/b"

    uri = os.path.join(path, "..", "test1.h5::/a/b")
    uriref = os.path.join(path, "test2.h5::/a")
    a, b = nexus.relUri(uri, uriref)
    assert a == "../test1.h5"
    assert b == "/a/b"


@pytest.mark.parametrize("abspath", [False, True])
def test_nexus_links(tmpdir, abspath):
    def namegen():
        i = 1
        while True:
            yield "link" + str(i)
            i += 1

    linkname = namegen()
    with nxroot(tmpdir, os.path.join("a", "b", "test1")) as f1:
        f1.create_group("a/b/c")
        g = f1["/a/b"]
        assert_same_target(g, g)
        # internal link up
        name = next(linkname)
        nexus.createLink(g, name, f1["a"], abspath=abspath)
        assert_same_target(f1["a"], g[name])
        link = g.get(name, getlink=True)
        assert link.path == "/a"
        assert isinstance(link, h5py.SoftLink)
        # internal link same level
        name = next(linkname)
        nexus.createLink(g, name, f1["a/b"], abspath=abspath)
        assert_same_target(f1["a/b"], g[name])
        link = g.get(name, getlink=True)
        assert link.path == "."
        assert isinstance(link, h5py.SoftLink)
        # internal link down
        name = next(linkname)
        nexus.createLink(g, name, f1["a/b/c"], abspath=abspath)
        assert_same_target(f1["a/b/c"], g[name])
        link = g.get(name, getlink=True)
        assert link.path == "c"
        assert isinstance(link, h5py.SoftLink)
        # external link down
        with nxroot(tmpdir, os.path.join("a", "test2")) as f2:
            name = next(linkname)
            nexus.createLink(f2, name, f1["a"], abspath=abspath)
            link = f2.get(name, getlink=True)
            assert_same_target(f1["a"], f2[name])
            assert link.path == "/a"
            if abspath:
                assert link.filename == f1.filename
            else:
                assert link.filename == "b/test1.h5"
            assert isinstance(link, h5py.ExternalLink)
        # internal link same level
        with nxroot(tmpdir, os.path.join("a", "b", "test2")) as f2:
            name = next(linkname)
            nexus.createLink(f2, name, f1["a"], abspath=abspath)
            link = f2.get(name, getlink=True)
            # assert_same_target(f1["a"], f2[name])
            assert link.path == "/a"
            if abspath:
                assert link.filename == f1.filename
            else:
                assert link.filename == "./test1.h5"
            assert isinstance(link, h5py.ExternalLink)
        # external link up
        with nxroot(tmpdir, os.path.join("a", "b", "c", "test2")) as f2:
            name = next(linkname)
            nexus.createLink(f2, name, f1["a"], abspath=abspath)
            assert_same_target(f1["a"], f2[name])
            link = f2.get(name, getlink=True)
            assert link.path, "/a"
            if abspath:
                assert link.filename == f1.filename
            else:
                assert link.filename == "../test1.h5"
            assert isinstance(link, h5py.ExternalLink)


def assert_same_target(node1, node2):
    target1 = nexus.dereference(node1)
    target2 = nexus.dereference(node2)
    assert nexus.normUri(target1) == nexus.normUri(target2)


def test_nexus_reshape_datasets(tmpdir):
    shape = 12, 5
    vshape = 3, 4, 5
    order = "C"

    def flatten(arr):
        return arr.flatten(order=order)

    kwargs = {
        "axis": 0,
        "virtual": True,
        "newaxis": False,
        "shape": vshape,
        "order": order,
        "fillvalue": 0,
    }
    fdatamem = numpy.arange(numpy.prod(shape))
    datamem = fdatamem.reshape(shape, order=order)

    def validate_dset(dset):
        numpy.testing.assert_array_equal(
            fdatamem, flatten(dset[()]), err_msg=nexus.getUri(dset)
        )

    filenames = (
        os.path.join("basedir1", "test1"),
        os.path.join("basedir1", "test2"),
        os.path.join("basedir1", "subdir", "test3"),
    )

    with nxroot(tmpdir, filenames[0]) as root1:
        with nxroot(tmpdir, filenames[1]) as root2:
            with nxroot(tmpdir, filenames[2]) as root3:
                for root in root1, root2, root3:
                    g = root.create_group("a")
                    g.create_group("b")
                    g["data"] = datamem
                    # Internal links
                    kwargs["data"] = [nexus.getUri(root["/a/data"])]
                    dset = nexus.nxCreateDataSet(root, "vdata", kwargs, None)
                    validate_dset(dset)
                    dset = nexus.nxCreateDataSet(root["/a"], "vdata", kwargs, None)
                    validate_dset(dset)
                    dset = nexus.nxCreateDataSet(root["/a/b"], "vdata", kwargs, None)
                    validate_dset(dset)

    # root1 -> root2, root3
    with nxroot(tmpdir, filenames[0]) as root1:
        kwargs["data"] = [nexus.getUri(root1["/a/data"])]

    for filename in filenames[1:]:
        with nxroot(tmpdir, filename) as root:
            dset = nexus.nxCreateDataSet(root, "vdatae", kwargs, None)
            validate_dset(dset)
        with nxroot(tmpdir, filename) as root:
            dset = nexus.nxCreateDataSet(root["/a"], "vdatae", kwargs, None)
            validate_dset(dset)
        with nxroot(tmpdir, filename) as root:
            dset = nexus.nxCreateDataSet(root["/a/b"], "vdatae", kwargs, None)
            validate_dset(dset)

    # root2 -> root1
    with nxroot(tmpdir, filenames[1]) as root2:
        kwargs["data"] = [nexus.getUri(root2["/a/data"])]
    with nxroot(tmpdir, filenames[0]) as root1:
        dset = nexus.nxCreateDataSet(root1, "vdatae", kwargs, None)
        validate_dset(dset)
    with nxroot(tmpdir, filenames[0]) as root1:
        dset = nexus.nxCreateDataSet(root1["/a"], "vdatae", kwargs, None)
        validate_dset(dset)
    with nxroot(tmpdir, filenames[0]) as root1:
        dset = nexus.nxCreateDataSet(root1["/a/b"], "vdatae", kwargs, None)
        validate_dset(dset)

    paths = ("/vdata", "/vdatae", "/a/vdata", "/a/vdatae", "/a/b/vdata", "/a/b/vdatae")
    for filename in filenames:
        with nxroot(tmpdir, filename) as root:
            data = root["/a/data"]
            assert shape == data.shape
            validate_dset(data)
            for path in paths:
                vdata = root[path]
                assert vshape == vdata.shape
                validate_dset(vdata)

    dirname = str(tmpdir)
    os.rename(os.path.join(dirname, "basedir1"), os.path.join(dirname, "basedir2"))
    os.rename(
        os.path.join(dirname, "basedir2", "test2.h5"),
        os.path.join(dirname, "basedir2", "test2_.h5"),
    )
    os.rename(
        os.path.join(dirname, "basedir2", "subdir", "test3.h5"),
        os.path.join(dirname, "basedir2", "subdir", "test3_.h5"),
    )
    filenames = (
        os.path.join("basedir2", "test1"),
        os.path.join("basedir2", "test2_"),
        os.path.join("basedir2", "subdir", "test3_"),
    )
    lostlinks = [
        ("/vdatae", "/a/vdatae", "/a/b/vdatae"),
        tuple(),
        ("/vdatae", "/a/vdatae", "/a/b/vdatae"),
    ]
    for filename, lost in zip(filenames, lostlinks):
        with nxroot(tmpdir, filename) as root:
            data = root["/a/data"]
            assert shape == data.shape
            validate_dset(data)
            for path in paths:
                vdata = root[path]
                assert vshape == vdata.shape
                if path in lost:
                    with pytest.raises(AssertionError):
                        validate_dset(vdata)
                else:
                    validate_dset(vdata)


def validateNxRoot(h5group):
    attrs = {
        "NX_class",
        "creator",
        "creator_version",
        "HDF5_Version",
        "file_name",
        "file_time",
        # "file_update_time",
        "h5py_version",
        "custom_root_attr",
    }
    assert set(h5group.attrs.keys()) == attrs
    assert h5group.attrs["NX_class"] == "NXroot"
    assert h5group.name == "/"


def validateNxEntry(h5group):
    attrs = ["NX_class"]
    assert set(h5group.attrs.keys()) == set(attrs)
    files = ["start_time", "end_time"]
    assert set(h5group.keys()) == set(files)
    assert h5group.attrs["NX_class"] == "NXentry"
    assert h5group.parent.name == "/"


def validateNxProcess(h5group):
    attrs = ["NX_class"]
    assert set(h5group.attrs.keys()) == set(attrs)
    files = ["program", "configuration", "date", "results"]
    assert set(h5group.keys()) == set(files)
    assert h5group.attrs["NX_class"] == "NXprocess"
    assert h5group.parent.attrs["NX_class"] == "NXentry"
    validateNxNote(h5group["configuration"])
    validateNxCollection(h5group["results"])


def validateNxNote(h5group):
    attrs = ["NX_class"]
    assert set(h5group.attrs.keys()) == set(attrs)
    files = ["date", "data", "type"]
    assert set(h5group.keys()) == set(files)
    assert h5group.attrs["NX_class"] == "NXnote"


def validateNxCollection(h5group):
    attrs = ["NX_class"]
    assert set(h5group.attrs.keys()) == set(attrs)
    assert h5group.attrs["NX_class"] == "NXcollection"


def validateNxData(h5group, axes, signals):
    attrs = ["NX_class", "axes", "signal", "auxiliary_signals"]
    assert set(h5group.attrs.keys()) == set(attrs)
    files = list(next(iter(zip(*axes)))) + list(next(iter(zip(*signals))))
    assert set(h5group.keys()) == set(files)
    assert h5group.attrs["NX_class"] == "NXdata"


def check_string_types(tmpdir, attribute=True, raiseExtended=True):
    # Test following string literals
    sAsciiBytes = b"abc"
    sAsciiUnicode = "abc"
    sLatinBytes = b"\xe423"
    sUTF8Unicode = "\u0101bc"
    sUTF8Bytes = b"\xc4\x81bc"
    sUTF8AsciiUnicode = "abc"
    sUTF8AsciiBytes = b"abc"
    # Expected conversion after HDF5 write/read
    strmap = {}
    strmap["ascii(scalar)"] = sAsciiBytes, sAsciiUnicode
    strmap["ext(scalar)"] = sLatinBytes, sLatinBytes
    strmap["unicode(scalar)"] = sUTF8Unicode, sUTF8Unicode
    strmap["unicode2(scalar)"] = sUTF8AsciiUnicode, sUTF8AsciiUnicode
    strmap["ascii(list)"] = [sAsciiBytes, sAsciiBytes], [sAsciiUnicode, sAsciiUnicode]
    strmap["ext(list)"] = [sLatinBytes, sLatinBytes], [sLatinBytes, sLatinBytes]
    strmap["unicode(list)"] = [sUTF8Unicode, sUTF8Unicode], [sUTF8Unicode, sUTF8Unicode]
    strmap["unicode2(list)"] = (
        [sUTF8AsciiUnicode, sUTF8AsciiUnicode],
        [sUTF8AsciiUnicode, sUTF8AsciiUnicode],
    )
    strmap["mixed(list)"] = (
        [sUTF8Unicode, sUTF8AsciiUnicode, sAsciiBytes, sLatinBytes],
        [sUTF8Bytes, sUTF8AsciiBytes, sAsciiBytes, sLatinBytes],
    )
    strmap["ascii(0d-array)"] = numpy.array(sAsciiBytes), sAsciiUnicode
    strmap["ext(0d-array)"] = numpy.array(sLatinBytes), sLatinBytes
    strmap["unicode(0d-array)"] = numpy.array(sUTF8Unicode), sUTF8Unicode
    strmap["unicode2(0d-array)"] = numpy.array(sUTF8AsciiUnicode), sUTF8AsciiUnicode
    strmap["ascii(1d-array)"] = (
        numpy.array([sAsciiBytes, sAsciiBytes]),
        [sAsciiUnicode, sAsciiUnicode],
    )
    strmap["ext(1d-array)"] = (
        numpy.array([sLatinBytes, sLatinBytes]),
        [sLatinBytes, sLatinBytes],
    )
    strmap["unicode(1d-array)"] = (
        numpy.array([sUTF8Unicode, sUTF8Unicode]),
        [sUTF8Unicode, sUTF8Unicode],
    )
    strmap["unicode2(1d-array)"] = (
        numpy.array([sUTF8AsciiUnicode, sUTF8AsciiUnicode]),
        [sUTF8AsciiUnicode, sUTF8AsciiUnicode],
    )
    strmap["mixed(1d-array)"] = (
        numpy.array([sUTF8Unicode, sUTF8AsciiUnicode, sAsciiBytes]),
        [sUTF8Unicode, sUTF8AsciiUnicode, sAsciiUnicode],
    )
    strmap["mixed2(1d-array)"] = (
        numpy.array([sUTF8AsciiUnicode, sAsciiBytes]),
        [sUTF8AsciiUnicode, sAsciiUnicode],
    )

    with nxroot(tmpdir, "test_nexus_String{:d}".format(attribute)) as h5group:
        h5group = h5group.create_group("test")
        if attribute:
            out = h5group.attrs
        else:
            out = h5group
        for name, (value, expectedValue) in strmap.items():
            decodingError = "ext" in name or name == "mixed(list)"
            if raiseExtended and decodingError:
                with pytest.raises(UnicodeDecodeError):
                    ovalue = nexus.asNxChar(value, raiseExtended=raiseExtended)
                continue
            else:
                ovalue = nexus.asNxChar(value, raiseExtended=raiseExtended)
            # Write/read
            out[name] = ovalue
            if attribute:
                value = h5py_read_attribute(out, name)
            else:
                value = h5py_read_dataset(out[name])
            # Expected type and value?
            if "list" in name or "1d-array" in name:
                assert isinstance(value, numpy.ndarray)
                value = value.tolist()
                assert list(map(type, value)) == list(map(type, expectedValue)), name
                firstValue = value[0]
            else:
                firstValue = value
            msg = "{} {} instead of {}".format(name, type(value), type(expectedValue))
            assert type(value) is type(expectedValue), msg
            assert value == expectedValue, msg
            # Expected character set?
            if not attribute:
                charSet = out[name].id.get_type().get_cset()
                if isinstance(firstValue, bytes):
                    # This is the tricky part, CSET_ASCII is supposed to be
                    # only 0-127 while we actually allow 0-255
                    expectedCharSet = h5py.h5t.CSET_ASCII
                else:
                    expectedCharSet = h5py.h5t.CSET_UTF8
                msg = "{} type {} instead of {}".format(name, charSet, expectedCharSet)
                assert charSet == expectedCharSet, msg


def test_nexus_exists(tmpdir):
    with nxroot(tmpdir, "test_nexus_entry") as root:
        uri = nexus.getUri(root)
    assert nexus.exists(uri)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        uri = nexus.hdf5_join(nexus.getUri(root), "entry")
    assert not nexus.exists(uri)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        nexus.nxEntry(root, "entry")
    assert nexus.exists(uri)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        with pytest.raises(nexus.NexusInstanceExists):
            nexus.nxEntry(root, "entry", raise_on_exists=True)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry"]
        uri = nexus.hdf5_join(nexus.getUri(parent), "collection")
    assert not nexus.exists(uri)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry"]
        nexus.nxCollection(parent, "collection")
    assert nexus.exists(uri)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry"]
        with pytest.raises(nexus.NexusInstanceExists):
            nexus.nxCollection(parent, "collection", raise_on_exists=True)
        # with pytest.raises(RuntimeError):
        nexus.nxCollection(root, "collection")

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry"]
        uri = nexus.hdf5_join(nexus.getUri(parent), "process")
    assert not nexus.exists(uri)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry"]
        nexus.nxProcess(parent, "process")
    assert nexus.exists(uri)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry"]
        with pytest.raises(nexus.NexusInstanceExists):
            nexus.nxProcess(parent, "process", raise_on_exists=True)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry"]
        uri = nexus.hdf5_join(nexus.getUri(parent), "subentry")
    assert not nexus.exists(uri)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry"]
        nexus.nxSubEntry(parent, "subentry")
    assert nexus.exists(uri)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry"]
        with pytest.raises(nexus.NexusInstanceExists):
            nexus.nxSubEntry(parent, "subentry", raise_on_exists=True)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry"]
        uri = nexus.hdf5_join(nexus.getUri(parent), "plot")
    assert not nexus.exists(uri)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry"]
        nexus.nxData(parent, "plot")
    assert nexus.exists(uri)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry"]
        with pytest.raises(nexus.NexusInstanceExists):
            nexus.nxData(parent, "plot", raise_on_exists=True)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry"]
        uri = nexus.hdf5_join(nexus.getUri(parent), "instrument")
    assert not nexus.exists(uri)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry"]
        nexus.nxInstrument(parent, "instrument")
    assert nexus.exists(uri)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry"]
        with pytest.raises(nexus.NexusInstanceExists):
            nexus.nxInstrument(parent, "instrument", raise_on_exists=True)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry/instrument"]
        uri = nexus.hdf5_join(nexus.getUri(parent), "detector")
    assert not nexus.exists(uri)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry/instrument"]
        nexus.nxDetector(parent, "detector")
    assert nexus.exists(uri)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry/instrument"]
        with pytest.raises(nexus.NexusInstanceExists):
            nexus.nxDetector(parent, "detector", raise_on_exists=True)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry/instrument"]
        uri = nexus.hdf5_join(nexus.getUri(parent), "positioner")
    assert not nexus.exists(uri)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry/instrument"]
        nexus.nxPositioner(parent, "positioner")
    assert nexus.exists(uri)

    with nxroot(tmpdir, "test_nexus_entry") as root:
        parent = root["entry/instrument"]
        with pytest.raises(nexus.NexusInstanceExists):
            nexus.nxPositioner(parent, "positioner", raise_on_exists=True)


def test_nexus_locked_exception(tmpdir):
    filename = os.path.join(str(tmpdir), "test.h5")

    def locking_process(q):
        import threading

        with nexus.File(filename, mode="a", locking=True) as f:
            f["a"] = 10
            f.flush()
            q.put(None)
            threading.Event().wait()

    q = multiprocessing.Queue()
    p = multiprocessing.Process(target=locking_process, args=(q,))
    p.start()
    try:
        # Wait until file is expected to be locked
        q.get(timeout=10)
        # We should not be able to lock the file
        try:
            with nexus.File(filename, mode="r", locking=True):
                pass
            raise RuntimeError("Locked exception not raises")
        except Exception as e:
            assert nexus.isLockedError(e)
            assert f"pid={p.pid}" in nexus.lockedErrorMessage(filename)
        # We should be able to open the file without locking
        with nexus.File(filename) as f:
            assert f["a"][()] == 10
    finally:
        p.kill()


def test_nexus_locked_self(tmpdir):
    filename = str(tmpdir / "test.h5")
    with nexus.File(filename, mode="a", locking=True) as f1:
        f1["a"] = 10
        f1.flush()
        # We should be able to open the file with locking
        with nexus.File(filename, mode="r", locking=True) as f2:
            assert f2["a"][()] == 10


def test_dicttonx_structured_array(tmpdir):
    filename = str(tmpdir / "test.h5")
    data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
    with nexus.File(filename, mode="w") as f:
        tree = {"empty": [], "data": data}
        nexus.dicttonx(tree, f)
    content = dictdump.nxtodict(filename)
    structures = content["data"]
    structures = [dict(zip(structures.dtype.names, x)) for x in structures]
    assert structures == data



import os
import enum
import logging
import datetime
import threading
from typing import Optional, Generator
from contextlib import contextmanager

from ..parameters.base import PROFILE_PARAMETERS
from ..utils import profiling
from ..utils import logging_utils

_logger = logging.getLogger(__name__)


class BaseSubscriber:
    """Provides state, logging, timing, sorting and concurrency model"""

    @enum.unique
    class _STATE(enum.IntEnum):
        INIT = enum.auto()
        ON = enum.auto()
        FAULT = enum.auto()
        OFF = enum.auto()

    _ALLOWED_STATE_TRANSITIONS = {
        _STATE.INIT: [_STATE.INIT, _STATE.ON, _STATE.OFF, _STATE.FAULT],
        _STATE.ON: [_STATE.OFF, _STATE.FAULT],
        _STATE.OFF: [_STATE.OFF, _STATE.FAULT],
        _STATE.FAULT: [],
    }

    def __init__(
        self,
        name: str,
        parent_logger=None,
        resource_profiling: Optional[PROFILE_PARAMETERS] = None,
        start_semaphore_file: Optional[str] = None,
    ) -> None:
        self._name = name
        self._state = self._STATE.INIT
        self._state_reason = "instantiation"
        self._init_time = datetime.datetime.now().astimezone()

        if parent_logger is None:
            parent_logger = _logger
        self._logger = logging_utils.CustomLogger(parent_logger, self)

        if resource_profiling is None:
            resource_profiling = PROFILE_PARAMETERS.OFF
        self.resource_profiling = resource_profiling

        self._thread: Optional[threading.Thread] = None
        self._started_event = threading.Event()
        self._start_time = None
        self._start_semaphore_file = start_semaphore_file
        self._stop_requested: bool = False
        self._end_time = None

    def __str__(self) -> str:
        return f"{self.name} ({self.state.name})"

    @property
    def name(self):
        return self._name

    @property
    def state(self) -> enum.IntEnum:
        return self._state

    @property
    def state_reason(self) -> str:
        return self._state_reason

    def _set_state(self, state: enum.IntEnum, reason: str, force: bool = False) -> None:
        if force or state in self._ALLOWED_STATE_TRANSITIONS[self._state]:
            self._state = self._STATE(state)
            self._state_reason = reason
            if state == self._STATE.FAULT:
                self._logger.error(reason)
            else:
                self._logger.info(reason)

    def _clear_state(self) -> None:
        self._stop_requested = False
        self._started_event.clear()
        self._start_time = None
        self._end_time = None
        self._set_state(self._STATE.INIT, "Starting", force=True)

    def start(self, block: bool = True, timeout: Optional[float] = None) -> None:
        if self.is_alive():
            return
        self._thread = threading.Thread(target=self._main)
        self._clear_state()
        self._thread.start()
        if block:
            self._started_event.wait(timeout=timeout)

    def stop(self, block: bool = True, timeout: Optional[float] = None) -> None:
        self._stop_requested = True
        if block:
            self.join(timeout=timeout)

    def join(self, timeout: Optional[float] = None) -> None:
        if self._thread:
            self._thread.join(timeout=timeout)

    def is_alive(self) -> bool:
        return self._thread and self._thread.is_alive()

    def _main(self) -> None:
        try:
            self._on_started()
            with self._possible_profiling():
                self._run()
        except KeyboardInterrupt:
            self._set_state(self._STATE.FAULT, "KeyboardInterrupt")
            self._logger.warning("Stop listening to Redis events (KeyboardInterrupt)")
            raise
        except BaseException as e:
            self._set_state(self._STATE.FAULT, str(e))
            self._logger.exception("Stop listening to Redis events due to an exception")
            raise
        finally:
            self.on_finished()

    def _on_started(self) -> None:
        self._started_event.set()
        self._start_time = self._now
        if self._start_semaphore_file:
            dirname = os.path.dirname(self._start_semaphore_file)
            if dirname:
                os.makedirs(dirname, exist_ok=True)
            with open(self._start_semaphore_file, "a"):
                pass

    def on_finished(self) -> None:
        try:
            if self._start_semaphore_file:
                os.unlink(self._start_semaphore_file)
        finally:
            self._end_time = self._now
            self._set_state(self._STATE.OFF, "Finished succesfully")

    def _run(self) -> None:
        self._set_state(self._STATE.ON, "Subscribed to Redis")

    @property
    def resource_profiling(self) -> PROFILE_PARAMETERS:
        return self._resource_profiling

    @resource_profiling.setter
    def resource_profiling(self, value: Optional[PROFILE_PARAMETERS]):
        if value is None:
            self._resource_profiling = PROFILE_PARAMETERS.OFF
        else:
            self._resource_profiling = PROFILE_PARAMETERS(value)

    @contextmanager
    def _possible_profiling(self) -> Generator[None, None, None]:
        kwargs = self.resource_profiling.arguments
        if kwargs:
            with profiling.profile(**kwargs):
                yield
        else:
            yield

    @property
    def init_time(self) -> Optional[datetime.datetime]:
        """Time of instatiation"""
        return self._init_time

    @property
    def start_time(self) -> Optional[datetime.datetime]:
        """Start of writing"""
        return self._start_time

    @property
    def end_time(self) -> Optional[datetime.datetime]:
        """End of writing"""
        return self._end_time

    @property
    def _now(self) -> datetime.datetime:
        return datetime.datetime.now().astimezone()

    def get_duration(self) -> datetime.timedelta:
        """Time between start and end of writing"""
        t0 = self._start_time
        t1 = self._end_time
        if t0 is None:
            t0 = self._init_time
        if t1 is None:
            t1 = self._now
        if t1 < t0:
            return t0 - t0
        else:
            return t1 - t0

    def _sort_key(self):
        if self._start_time is None:
            return self._init_time
        else:
            return self._start_time

    def __lt__(self, other):
        if isinstance(other, type(self)):
            return self._sort_key() < other._sort_key()



import os
from typing import Optional


def find_existing(path: str) -> Optional[str]:
    """Returns `path` or one of its parent directories."""
    path = os.path.normpath(path)
    while not os.path.exists(path):
        previous = path
        path = os.path.dirname(path)
        if path == previous:
            break
    if not os.path.exists(path):
        return
    return path


def has_required_disk_space(
    path: str,
    required_disk_space: float,
    query_pool=None,
) -> bool:
    """
    :param path: may not exist yet
    :param required_disk_space: is MB
    :param query_pool:
    :returns: also returns `True` when no path was found or
              the call did not finish within the query pool's
              timeout.
    """
    if required_disk_space <= 0:
        return True
    path = find_existing(path)
    if not path:
        return True
    stat = statvfs(path, query_pool=query_pool)
    if stat is None:
        return True
    free_space = stat.f_frsize * stat.f_bavail / 1024**2
    return free_space >= required_disk_space


def statvfs(path, query_pool=None):
    """os.statvfs could take several seconds on NFS"""
    if query_pool is None:
        return os.statvfs(path)
    return query_pool.execute(os.statvfs, args=(path,), default=None)


def has_write_permissions(path: str) -> bool:
    if os.path.exists(path):
        return os.access(path, os.W_OK)
    # Check whether we can create the path
    path = os.path.dirname(os.path.normpath(path))
    path = find_existing(path)
    if path and os.path.isdir(path):
        return os.access(path, os.W_OK)
    return False



"""Static Bliss configuration utilities
"""

import os


def static_config():
    """
    :returns bliss.config.static.Config:
    """
    from bliss.config import static  # avoid bliss patching on import

    return static.get_config()


def static_root(root=None):
    """
    :returns ConfigNode:
    """
    if root is None:
        return static_config().root
    else:
        return root


def static_root_find(name, root=None):
    """
    :returns ConfigNode:
    """
    from bliss.config import static  # avoid bliss patching on import

    root = static_root(root=root)
    if root.children:
        for node in root.children:
            if node.get("name", None) == name:
                return node
    nodes = []
    for node in root.values():
        if isinstance(node, static.ConfigNode):
            nodes.append(node)
        elif isinstance(node, list):
            for nodei in node:
                if isinstance(nodei, static.ConfigNode):
                    nodes.append(nodei)
    for node in nodes:
        if node.get("name", None) == name:
            return node
        ret = static_root_find(name, root=node)
        if ret is not None:
            return ret
    return None


def beamline(root=None, default="id00"):
    """
    :returns str:
    """
    name = default
    for k in "BEAMLINENAME", "BEAMLINE":
        name = os.environ.get(k, name)
    root = static_root(root=root)
    name = root.get("beamline", name)
    scan_saving = static_root_find("scan_saving", root=root)
    if scan_saving is not None:
        name = scan_saving.get("beamline", name)
    return name.lower()


def institute(root=None, default=""):
    """
    :returns str:
    """
    root = static_root(root=root)
    name = default
    name = root.get("institute", name)
    name = root.get("laboratory", name)
    name = root.get("synchrotron", name)
    return name


def instrument(root=None, default=""):
    """
    :returns str:
    """
    root = static_root(root=root)
    default_instrument = institute(root=root, default=default)
    if default and default.lower() not in default_instrument.lower():
        default_instrument = f"{default_instrument}-{default}"
    name = root.get("instrument", default_instrument)
    return name



import os
from io import StringIO
from contextlib import contextmanager, ExitStack
from time import perf_counter

try:
    import tracemalloc
    import linecache
except ImportError:
    tracemalloc = None
import cProfile
import pstats

try:
    import yappi
except ImportError:
    yappi = None
else:
    try:
        from gevent.monkey import is_anything_patched
    except ImportError:

        def is_anything_patched():
            return False

    if is_anything_patched():
        yappi.set_context_backend("greenlet")

from .logging_utils import log
from ..io import io_utils


DEFAULT_FILENAME = os.path.join(
    io_utils.temproot(), "pyprof_pid{}.cprof".format(os.getpid())
)


class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def durationfmtcolor(x):
    if x < 0.0005:
        return bcolors.OKGREEN + ("%6dµs" % (x * 1000000)) + bcolors.ENDC
    elif x < 0.001:
        return bcolors.WARNING + ("%6dµs" % (x * 1000000)) + bcolors.ENDC
    elif x < 0.1:
        return bcolors.WARNING + ("%6dms" % (x * 1000)) + bcolors.ENDC
    else:
        return bcolors.FAIL + ("%8.3f" % x) + bcolors.ENDC


def durationfmt(x):
    if x < 0.001:
        return "%6dµs" % (x * 1000000)
    elif x < 0.1:
        return "%6dms" % (x * 1000)
    else:
        return "%8.3f" % x


@contextmanager
def execute_min_duration(func, min_duration=None):
    """Execute function when the time >= min_duration.

    :param callable func:
    :param num min_duration:
    """
    t0 = perf_counter()
    try:
        yield
    finally:
        if not min_duration or (perf_counter() - t0) >= min_duration:
            func()


def log_malloc_snapshot(snapshot, logger=None, key_type="lineno", limit=10, units="KB"):
    """
    :param tracemalloc.Snapshot snapshot:
    :param str key_type:
    :param int limit: limit number of lines
    :param str units: B, KB, MB, GB
    :param logger:
    """
    n = ["b", "kb", "mb", "gb"].index(units.lower())
    sunits, units = units, 1024**n

    snapshot = snapshot.filter_traces(
        (
            tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
            tracemalloc.Filter(False, "<unknown>"),
        )
    )
    top_stats = snapshot.statistics(key_type)
    total = sum(stat.size for stat in top_stats)

    log(logger, "================Memory profile================")
    out = ""
    for index, stat in enumerate(top_stats, 1):
        frame = stat.traceback[0]
        # replace "/path/to/module/file.py" with "module/file.py"
        # filename = os.sep.join(frame.filename.split(os.sep)[-2:])
        filename = frame.filename
        out += "\n#%s: %s:%s: %.1f %s" % (
            index,
            filename,
            frame.lineno,
            stat.size / units,
            sunits,
        )
        line = linecache.getline(frame.filename, frame.lineno).strip()
        if line:
            out += "\n    %s" % line
        if index >= limit:
            break
    other = top_stats[index:]
    if other:
        size = sum(stat.size for stat in other)
        out += "\n%s other: %.1f %s" % (len(other), size / units, sunits)
    out += "\nTotal allocated size: %.1f %s" % (total / units, sunits)
    log(logger, out)
    log(logger, "============================================")


DEFAULT_SORTBY = "tottime"
DEFAULT_CLOCK = "cpu"


def log_pstats_snapshot(
    snapshot, logger=None, timelimit=None, sortby=None, color=False
):
    """
    :param Stats snapshot:
    :param logger:
    :param int or float timelimit: number of lines or fraction (float between 0 and 1)
    :param str sortby: sort time profile
    :param bool color:
    """
    if isinstance(sortby, str):
        sortby = [sortby]
    elif not sortby:
        sortby = [DEFAULT_SORTBY]
    for sortmethod in sortby:
        snapshot.stream = s = StringIO()
        if color:
            pstats.f8 = durationfmtcolor
        else:
            pstats.f8 = durationfmt
        ps = snapshot.sort_stats(sortmethod)
        if timelimit is None:
            timelimit = tuple()
        elif not isinstance(timelimit, tuple):
            timelimit = (timelimit,)
        ps.print_stats(*timelimit)

        log(logger, "================Time profile================")
        msg = "\n" + s.getvalue()
        log(logger, msg)
        log(logger, "============================================")


pstats_to_yappi_sort = {
    None: DEFAULT_SORTBY,
    "name": "name",
    "ncalls": "ncall",
    "cumtime": "ttot",  # including subcalls
    "tottime": "tsub",  # exluding subcalls
}


def log_yappi_snapshot(snapshot, logger=None, sortby=None):
    """
    :param YFuncStat snapshot:
    :param logger:
    :param str sortby: sort time profile
    """
    if isinstance(sortby, str):
        sortby = [sortby]
    elif not sortby:
        sortby = [DEFAULT_SORTBY]
    for sortmethod in sortby:
        s = StringIO()
        sortmethod = pstats_to_yappi_sort[sortmethod]
        snapshot = snapshot.sort(sortmethod)
        snapshot.print_all(out=s)
        log(logger, "================Time profile================")
        msg = "\n" + s.getvalue()
        log(logger, msg)
        log(logger, "============================================")


@contextmanager
def memory_context(logger=None, min_duration=None, **kwargs):
    """
    :param logger:
    :param **kwargs: see log_malloc_snapshot
    """
    if tracemalloc is None:
        log(logger, "tracemalloc required")
        return
    tracemalloc.clear_traces()
    tracemalloc.start()

    def func():
        snapshot = tracemalloc.take_snapshot()
        log_malloc_snapshot(snapshot, logger=logger, **kwargs)

    try:
        with execute_min_duration(func, min_duration=min_duration):
            yield
    finally:
        tracemalloc.stop()


@contextmanager
def time_context_cprofile(
    logger=None,
    timelimit=None,
    clock=None,
    sortby=None,
    color=False,
    filename=None,
    min_duration=None,
):
    """
    :param logger:
    :param int or float timelimit: number of lines or fraction (float between 0 and 1)
    :param str clock: always wall
    :param str sortby: sort time profile
    :param bool color:
    :param str or bool filename:
    :param num min_duration: no output when time < min_duration
    """
    pr = cProfile.Profile()
    pr.enable()

    def func():
        nonlocal filename
        snapshot = pstats.Stats(pr)
        log_pstats_snapshot(
            snapshot, logger=logger, timelimit=timelimit, sortby=sortby, color=color
        )
        if filename:
            if not isinstance(filename, str):
                filename = DEFAULT_FILENAME
            io_utils.rotatefiles(filename)
            # for pyprof2calltree
            snapshot.dump_stats(filename)
            log(logger, f"Statistics saved as {repr(filename)}")

    with execute_min_duration(func, min_duration=min_duration):
        try:
            yield
        finally:
            pr.disable()


@contextmanager
def time_context_yappi(
    logger=None,
    timelimit=None,
    clock=None,
    sortby=None,
    color=False,
    filename=None,
    min_duration=None,
    yappi_output=False,
):
    """
    :param logger:
    :param int or float timelimit: number of lines or fraction (float between 0 and 1)
    :param str clock: "wall" or "cpu"
    :param str sortby: sort time profile
    :param bool color:
    :param str or bool filename:
    :param num min_duration: no output when time < min_duration
    :param bool yappi_output: output formatted by yappi
    """
    if clock not in ("cpu", "wall"):
        clock = DEFAULT_CLOCK
    yappi.set_clock_type(clock)
    yappi.clear_stats()
    yappi.start(builtins=False)

    def func():
        nonlocal filename
        stats = yappi.get_func_stats()
        yappi.clear_stats()
        if yappi_output:
            log_yappi_snapshot(stats)
        else:
            snapshot = yappi.convert2pstats(stats)
            log_pstats_snapshot(
                snapshot, logger=logger, timelimit=timelimit, sortby=sortby, color=color
            )
        if filename:
            if not isinstance(filename, str):
                filename = DEFAULT_FILENAME
            io_utils.rotatefiles(filename)
            # callgrind: for qcachegrind
            stats.save(filename, type="callgrind")
            # for pyprof2calltree
            # snapshot.save(filename)
            log(logger, f"Statistics saved as {repr(filename)}")

    with execute_min_duration(func, min_duration=min_duration):
        try:
            yield
        finally:
            yappi.stop()


if yappi is None:
    time_context = time_context_cprofile
else:
    time_context = time_context_yappi


@contextmanager
def profile_context(
    memory=False,
    time=True,
    memlimit=10,
    timelimit=None,
    clock=None,
    sortby=None,
    color=False,
    filename=None,
    units="KB",
    logger=None,
    min_duration=None,
):
    """
    :param bool memory: profile memory usage
    :param bool time: execution time
    :param int memlimit: number of lines
    :param int or float timelimit: number of lines or fraction (float between 0 and 1)
    :param str clock: "wall" or "cpu"
    :param str sortby: sort time profile
    :param bool color:
    :param str or bool filename: dump for visual tools
    :param str units: memory units
    :param logger:
    :param num min_duration: no output when time < min_duration
    """
    with ExitStack() as stack:
        if time:
            ctx = time_context(
                timelimit=timelimit,
                clock=clock,
                sortby=sortby,
                color=color,
                filename=filename,
                logger=logger,
                min_duration=min_duration,
            )
            stack.enter_context(ctx)
        if memory:
            ctx = memory_context(
                limit=memlimit, units=units, logger=logger, min_duration=min_duration
            )
            stack.enter_context(ctx)
        yield


class ProfilerMeta(type):
    """Singleton pattern but with updating profiler arguments"""

    _instance = None

    def __call__(cls, **kw):
        if cls._instance is None:
            cls._instance = super(ProfilerMeta, cls).__call__(**kw)
        else:
            cls._instance._kw = kw
        return cls._instance


class profile(metaclass=ProfilerMeta):
    """Singleton profile manager for time and memory profiling"""

    def __init__(self, **kw):
        self._kw = kw
        self._ctr = 0
        self._ctx = None

    def __enter__(self):
        self._ctr += 1
        if self._ctx is None:
            self._ctx = profile_context(**self._kw)
            self._ctx.__enter__()

    def __exit__(self, *args):
        self._ctr -= 1
        if not self._ctr:
            self._ctx.__exit__(*args)
            self._ctx = None



"""Writer configuration to be published in Redis
"""

import logging

# TODO: should not import bliss in this project
from bliss.common.counter import SamplingMode
from bliss.common.counter import SamplingCounter

from bliss.common.auto_filter.counters import AutoFilterCalcCounter

from bliss.controllers.mca.counter import SpectrumMcaCounter as McaSpectrum
from bliss.controllers.mca.counter import StatisticsMcaCounter as McaStat
from bliss.controllers.mca.counter import RoiMcaCounter as McaRoi

from bliss.controllers.mca.mythen import MythenCounter as MythenSpectrum
from bliss.controllers.mca.mythen import RoiMythenCounter as MythenRoi

from bliss.controllers.mosca.counters import SpectrumCounter as MoscaSpectrum
from bliss.controllers.mosca.counters import StatCounter as MoscaStat
from bliss.controllers.mosca.counters import ROICounter as MoscaRoi

from bliss.controllers.lima.bpm import LimaBpmCounter as LimaBpm
from bliss.controllers.lima.image import ImageCounter as LimaImage
from bliss.controllers.lima.roi import RoiStatCounter as LimaRoi
from bliss.controllers.lima.roi import RoiProfileCounter as LimaProfile
from bliss.controllers.lima.roi import RoiCollectionCounter as LimaCollection

from ..utils import config_utils
from ..utils import session_utils
from ..utils import scan_utils


logger = logging.getLogger(__name__)


CATEGORIES = ["NEXUSWRITER"]


def register_metadata_generators(user_scan_meta):
    """Create the metadata generators for the configurable writer

    :param bliss.scanning.scan_meta.ScanMeta user_scan_meta:
    """
    user_scan_meta.add_categories(CATEGORIES)
    nexuswriter = user_scan_meta.nexuswriter
    nexuswriter.set("instrument_info", fill_instrument_info)
    nexuswriter.set("device_info", fill_device_info)
    nexuswriter.set("technique", fill_technique_info)
    nexuswriter.set("masterfiles", fill_masterfiles)


def fill_instrument_info(scan):
    """
    :param bliss.scanning.scan.Scan scan:
    """
    logger.debug("fill instrument info")
    short_name = config_utils.beamline(default="")
    short_name = session_utils.scan_saving_get(
        "beamline", scan_saving=scan.scan_saving, default=short_name
    )
    name = config_utils.instrument(default=short_name)
    return {"instrument_info": {"name": name, "name@short_name": short_name}}


def fill_technique_info(scan):
    """
    :param bliss.scanning.scan.Scan scan:
    """
    logger.debug("fill technique info")
    return {"technique": current_technique_definition(scan.scan_saving)}


def fill_masterfiles(scan):
    """
    :param bliss.scanning.scan.Scan scan:
    """
    logger.debug("fill master filenames")
    if scan.scan_info["save"]:
        return {"masterfiles": scan_utils.session_master_filenames()}
    return dict()


def fill_device_info(scan):
    """
    :param bliss.scanning.scan.Scan scan:
    """
    logger.debug("fill device info")
    return {"devices": device_info(scan)}


def _mca_device_info(ctr):
    """
    :param BaseMcaCounter ctr:
    :returns str:
    """
    description = (
        ctr._counter_controller.detector_brand.name
        + "/"
        + ctr._counter_controller.detector_type.name
    )
    return {"type": "mca", "description": description}


def _samplingcounter_device_info(ctr):
    """
    :param SamplingCounter ctr:
    :returns str:
    """
    return {"type": "samplingcounter", "mode": ctr.mode.name}


def device_info(scan):
    """
    Publish information on devices (defines types and groups counters).
    Bliss has the concept of controllers and data nodes, but the intermediate
    device level is missing so we need to create it here.

    :param bliss.scanning.scan.Scan scan:
    :returns dict:
    """
    from bliss import global_map  # avoid bliss patching on import

    devices = {}
    for ctr in global_map.get_counters_iter():
        _device_info_add_ctr(devices, ctr)
    return devices


def _device_info_add_ctr(devices, ctr):
    """
    :param dict devices: str -> dict
    :param ctr:
    """
    from bliss import global_map  # avoid bliss patching on import

    try:
        fullname = ctr.fullname.replace(".", ":")  # Redis name
    except AttributeError:
        logger.info(
            "%s does not have a fullname (most likely not a channel)", repr(ctr)
        )
        return
    alias = global_map.aliases.get_alias(ctr)
    if isinstance(ctr, MythenSpectrum):
        device_info = {"type": "mythen"}
        device = {"device_info": device_info, "device_type": "mythen"}
        devices[fullname] = device
    elif isinstance(ctr, MythenRoi):
        device_info = {"type": "mythen"}
        device = {"device_info": device_info, "device_type": "mythen"}
        devices[fullname] = device
    elif isinstance(ctr, McaSpectrum):
        device_info = {"type": "mca"}
        device = {"device_info": device_info, "device_type": "mca"}
        devices[fullname] = device
    elif isinstance(ctr, McaStat):
        device_info = {"type": "mca"}
        device = {"device_info": device_info, "device_type": "mca"}
        devices[fullname] = device
    elif isinstance(ctr, McaRoi):
        device_info = {"type": "mca"}
        device = {"device_info": device_info, "device_type": "mca"}
        devices[fullname] = device
    elif isinstance(ctr, MoscaSpectrum):
        # "falconx:spectrum" -> "falconx:spectrum:det_00", "falconx:spectrum:det_01", ...
        device_info = {"type": "mca"}
        device = {"device_info": device_info, "device_type": "mosca"}
        for i in ctr._counter_controller._mca.active_channels.values():
            devices[f"{fullname}:det{i:02d}"] = device
    elif isinstance(ctr, MoscaStat):
        # "falconx:xxxxx" -> "falconx:stat:xxxxx_det00", "falconx:stat:xxxxx_det01", ...
        device_info = {"type": "mca"}
        device = {"device_info": device_info, "device_type": "mosca"}
        parts = fullname.split(":")
        statname = parts[-1]
        prefix = ":".join(parts[:-1])
        for i in ctr._counter_controller._mca.active_channels.values():
            devices[f"{prefix}:stat:{statname}_det{i:02d}"] = device
    elif isinstance(ctr, MoscaRoi):
        # "falconx:roi1" -> "falconx:roi:roi1"
        device_info = {"type": "mca"}
        device = {
            "device_info": device_info,
            "device_type": "mosca",
            "channel": ctr.roi.channel,
        }
        parts = fullname.split(":")
        parts = parts[:-1] + ["roi", parts[-1]]
        devices[":".join(parts)] = device
    elif isinstance(ctr, LimaBpm):
        device_info = {"type": "lima"}
        device = {"device_info": device_info, "device_type": "lima"}
        devices[fullname] = device
    elif isinstance(ctr, LimaImage):
        device_info = {"type": "lima"}
        device = {"device_info": device_info, "device_type": "lima"}
        devices[fullname] = device
    elif isinstance(ctr, (LimaRoi, LimaProfile, LimaCollection)):
        device_info = {"type": "lima"}
        device = {"device_info": device_info, "device_type": "lima"}
        devices[fullname] = device
    elif isinstance(ctr, AutoFilterCalcCounter):
        device_info = {"type": "autofilter"}
        device = {"device_info": device_info, "device_type": "autofilter"}
        devices[fullname] = device
    elif isinstance(ctr, SamplingCounter):
        device_info = _samplingcounter_device_info(ctr)
        device = {
            "device_info": device_info,
            "device_type": "samplingcounter",
            "data_type": "signal",
        }
        devices[fullname] = device
        if ctr.mode == SamplingMode.SAMPLES:
            device = {"device_info": device_info, "device_type": "samplingcounter"}
            devices[fullname + "_samples"] = device
            if alias:
                devices[fullname + "_samples"]["alias"] = alias + "_samples"
        elif ctr.mode == SamplingMode.STATS:
            for stat in "N", "std", "var", "min", "max", "p2v":
                device = {"device_info": device_info, "device_type": "samplingcounter"}
                devices[fullname + "_" + stat] = device
                if alias:
                    devices[fullname + "_" + stat]["alias"] = alias + "_" + stat
    else:
        logger.info(
            "Counter %s %s published as generic detector",
            fullname,
            type(ctr).__qualname__,
        )
        devices[fullname] = {}
    if alias:
        devices[fullname]["alias"] = alias


def writer_config():
    """
    Get writer configuration from the static session configuration

    :returns dict:
    """
    cfgnode = config_utils.static_root_find("nexus_definitions")
    if cfgnode is None:
        return dict()
    else:
        return cfgnode.to_dict()


def writer_config_get(name, default=None):
    """
    Get attribute from the writer configuration

    :returns str:
    """
    return writer_config().get(name, default)


def technique_info():
    """
    Information on techniques from the static writer configuration

    :returns dict:
    """
    return writer_config_get("technique", {})


def technique_info_get(name, default=None):
    """
    Get attribute from the technique info

    :returns str:
    """
    return technique_info().get(name, default)


def default_technique():
    """
    Default technique from the technique info

    :returns str:
    """
    return technique_info_get("default", "undefined")


def current_technique(scan_saving=None):
    """
    Active technique from the session's scan saving object

    :returns str:
    """
    return session_utils.dataset_get(
        "definition", default_technique(), scan_saving=scan_saving
    )


def techniques():
    """
    List of available techniques from the technique info

    :returns list:
    """
    return list(technique_info_get("techniques", {}).keys())


def technique_definition(technique):
    """
    Technique definition from the technique info

    :param str techique:
    :returns dict: {'name': str,
                    'applications': dict(dict),
                    'plots': dict(list),
                    'plotselect': str}
    """
    applications = {}
    plots = {}
    ret = {
        "name": technique,
        "applications": applications,
        "plots": plots,
        "plotselect": "",
    }
    info = technique_info_get("techniques", {}).get(technique, {})
    if info is None:
        info = {}

    # Get the application definitions selected for this technique
    applicationdict = technique_info_get("applications", {})
    for name in info.get("applications", []):
        definition = applicationdict.get(name, {})
        # for example {'xrf':{'I0': 'iodet',
        #                     'It': 'idet',
        #                     'mca': [...]}, ...}
        if definition:
            name = definition.pop("personal_name", name)
            applications[name] = definition

    # Get the plots selected for this technique
    plotdict = technique_info_get("plots", {})
    plotselect = ""  # this first one is the default
    for name in info.get("plots", []):
        plotdefinition = plotdict.get(name, {})
        # for examples:
        #   {'personal_name': 'counters', 'items': ['iodet', 'xmap1:deadtime_det2', ...]}
        #   {'personal_name': 'counters', 'ndim': 2, 'grid': True}
        items = plotdefinition.get("items", [])
        ndim = plotdefinition.get("ndim", -1)
        grid = plotdefinition.get("grid", False)
        if not items and ndim < 0:
            continue
        name = plotdefinition.get("personal_name", name)
        if name in applications:
            name = name + "_plot"
        if not plotselect:
            plotselect = name
        plots[name] = {"items": items, "grid": grid, "ndim": ndim}
    ret["plotselect"] = plotselect
    return ret


def current_technique_definition(scan_saving=None):
    """
    Current technique definition from the technique info

    :returns dict(dict): technique:definition (str:dict)
    """
    return technique_definition(current_technique(scan_saving=scan_saving))



import os
import random
import itertools
from typing import Set

import h5py
import numpy
import pytest
from fabio.edfimage import EdfImage

from blisswriter.io import nexus
from blisswriter.io import h5_config
from blisswriter.utils.array_order import Order
from blisswriter.writer.datasets import Hdf5DatasetWriter

from .helpers.nxw_test_math import asproduct


@pytest.mark.parametrize("scan_shape", (tuple(), (0,), (8,), (4, 8), (2, 4, 6)))
@pytest.mark.parametrize("detector_shape", (tuple(), (3,), (3, 5)))
@pytest.mark.parametrize("source_file_type", ("hdf5", "multi-hdf5", "edf", None))
def test_dataset_proxy_shapes_and_formats(
    scan_shape, detector_shape, source_file_type, tmpdir
):
    nframes_per_file = 3
    fulluri = source_file_type and "multi-" in source_file_type
    if fulluri:
        source_file_type = source_file_type.replace("multi-", "")

    mainfile = str(tmpdir.join("main.h5"))
    with nexus.nxRoot(mainfile, mode="w") as nxroot:
        nexus.nxEntry(nxroot, "entry0000")
    dsetname_generator = name_generator("dataset{:04d}")
    sourcedir_generator = directory_generator(tmpdir)

    scan_save_ndims = 0, 1, 2
    saveorders = (
        Order(order="C"),
        Order(order="F"),
        Order(order="C", caxes="all"),
        Order(order="F", caxes="all"),
    )
    publishorders = tuple(saveorders)
    nextras = 0, -2, 2

    options = [saveorders, publishorders, nextras, scan_save_ndims]
    for params in itertools.product(*options):
        saveorder, publishorder, nextra, scan_save_ndim = params
        if source_file_type == "edf" and saveorder.forder:
            continue

        # Number of points published: npoints + nextra
        if not detector_shape and source_file_type:
            continue
        if not scan_shape:
            if nextra > 0:
                continue
            elif nextra < 0:
                nextra = -1

        # Scan part of the saved dataset shape
        npoints = numpy.prod(scan_shape, dtype=int)
        if npoints:
            # Fixed length scan
            lst = list(asproduct(npoints, scan_save_ndim, includeone=False))
            if not lst:
                continue
            escan_save_shape = scan_save_shape = random.choice(lst)
        else:
            # Variable length scan
            scan_save_shape = scan_shape
            escan_save_shape = tuple(n if n else i for i, n in enumerate(scan_shape, 4))
            npoints = numpy.prod(escan_save_shape, dtype=int)

        # Expected shape after publish + reshape
        edetector_shape = detector_shape
        if saveorder.corder:
            escan_shape = escan_save_shape + edetector_shape
        else:
            escan_shape = edetector_shape + escan_save_shape
        if fulluri:
            external_images_per_file = nframes_per_file
        else:
            external_images_per_file = None

        err_msg = str(
            {
                "scan_shape": scan_shape,
                "scan_save_shape": scan_save_shape,
                "detector_shape": detector_shape,
                "saveorder": str(saveorder),
                "publishorder": str(publishorder),
                "nextra": nextra,
                "npoints": npoints,
            }
        )

        name = next(dsetname_generator)
        dproxy = Hdf5DatasetWriter(
            filename=mainfile,
            parent="/entry0000",
            device={"data_name": name, "data_alias": None, "data_info": {"units": "s"}},
            scan_shape=scan_shape,
            scan_save_shape=scan_save_shape,
            detector_shape=detector_shape,
            external_images_per_file=external_images_per_file,
            external_uri_from_file=not fulluri,
            dtype=float,
            saveorder=saveorder,
            publishorder=publishorder,
        )

        # Add data
        source_dir = next(sourcedir_generator)
        source_dir.mkdir()
        if source_file_type == "hdf5":
            data_generator = hdf5_data_generator(
                source_dir,
                "file",
                detector_shape,
                nframes=nframes_per_file,
                fulluri=fulluri,
            )
        elif source_file_type == "edf":
            data_generator = edf_data_generator(
                source_dir, "file", detector_shape, nframes=nframes_per_file
            )
        else:
            data_generator = mem_data_generator(detector_shape)
        add_data(dproxy, data_generator, npoints + nextra)

        if source_file_type == "hdf5" and fulluri:
            # Creating a VDS: no need to open the files
            os.chmod(str(source_dir), 0)

        # Make sure it exists with the expected shape
        if dproxy.is_external:
            dproxy.reshape(escan_save_shape)
            dproxy.flush()
        else:
            dproxy.flush()
            dproxy.reshape(escan_save_shape)
        assert dproxy.current_scan_save_shape == escan_save_shape, err_msg
        assert dproxy.current_shape == escan_shape, err_msg
        assert dproxy.current_detector_shape == edetector_shape, err_msg

        if source_file_type == "hdf5" and fulluri:
            # Allow access again to check the content
            os.chmod(str(source_dir), 0o755)

        _validate_data(dproxy, npoints, nextra, escan_save_shape, err_msg)


@pytest.fixture
def small_compression_limit():
    limit = h5_config.DEFAULT_COMPRESSION_LIMIT_NBYTES
    h5_config.DEFAULT_COMPRESSION_LIMIT_NBYTES = 0
    yield
    h5_config.DEFAULT_COMPRESSION_LIMIT_NBYTES = limit


@pytest.mark.parametrize(
    "compression_scheme",
    (
        None,
        "none",
        "gzip",
        "bitshuffle",
        "gzip-byteshuffle",
        "lz4-bitshuffle",
        "wrong_value",
    ),
)
def test_dataset_proxy_compression(compression_scheme, small_compression_limit, tmpdir):
    mainfile = str(tmpdir.join("main.h5"))
    with nexus.nxRoot(mainfile, mode="w") as nxroot:
        nexus.nxEntry(nxroot, "entry0000")

    npoints = 1
    detsize = 1024

    data = numpy.arange(npoints * detsize).reshape((npoints, detsize))
    chunk_options = {"compression_scheme": compression_scheme}

    if compression_scheme == "wrong_value":
        with pytest.raises(ValueError):
            dproxy = Hdf5DatasetWriter(
                filename=mainfile,
                parent="/entry0000",
                device={"data_name": "data", "data_alias": None},
                scan_shape=(0,),
                detector_shape=(detsize,),
                dtype=data.dtype,
                chunk_options=chunk_options,
            )
        return

    dproxy = Hdf5DatasetWriter(
        filename=mainfile,
        parent="/entry0000",
        device={"data_name": "data", "data_alias": None},
        scan_shape=(0,),
        detector_shape=(detsize,),
        dtype=data.dtype,
        chunk_options=chunk_options,
    )
    dproxy.add(data)
    dproxy.flush()  # flush chunk buffer

    with nexus.nxRoot(mainfile, mode="r") as nxroot:
        dset = nxroot["/entry0000/data"]
        compression = hdf5_compression_name(dset)
        if compression_scheme in (None, "gzip-byteshuffle"):
            assert compression == {"deflate", "shuffle"}
        elif compression_scheme == "none":
            assert not compression
        elif compression_scheme == "gzip":
            assert compression == {"deflate"}
        elif compression_scheme == "bitshuffle":
            assert compression == {
                "bitshuffle; see https://github.com/kiyo-masui/bitshuffle"
            }
        elif compression_scheme == "lz4-bitshuffle":
            assert compression == {
                "bitshuffle; see https://github.com/kiyo-masui/bitshuffle"
            }
        else:
            raise ValueError(compression_scheme)
        numpy.testing.assert_array_equal(data, dset[()])


def directory_generator(tmpdir):
    i = 1
    while True:
        yield tmpdir.join(f"data{i}")
        i += 1


def _validate_data(dproxy, npoints, nextra, escan_save_shape, err_msg):
    if dproxy.csaveorder:
        detector_save_axes = tuple(
            range(dproxy.scan_save_ndim, dproxy.scan_save_ndim + dproxy.detector_ndim)
        )
    else:
        detector_save_axes = tuple(range(dproxy.detector_ndim))

    with dproxy.open() as dset:
        data = dset[()].max(axis=detector_save_axes)
        expected = numpy.arange(1, npoints + 1, dtype=dproxy.dtype)
        if nextra < 0:
            expected[nextra:] = dproxy.fillvalue
        if escan_save_shape:
            expected = dproxy.saveorder.reshape(expected, escan_save_shape)
        numpy.testing.assert_array_equal(data, expected, err_msg=err_msg)


def add_data(dproxy, data_generator, npoints):
    newdata = []
    for i in range(npoints):
        newdata.append(next(data_generator))
        if random.choice([0, 0, 1]):
            if isinstance(newdata[0], tuple):
                filename, index = newdata[0]
                filename = nexus.splitUri(filename)[0]
                file_format = os.path.splitext(filename)[-1][1:]
                dproxy.add_external(newdata, file_format=file_format)
            else:
                newdata = numpy.array(newdata)
                dproxy.add_internal(newdata)
            newdata = []
    if newdata:
        if isinstance(newdata[0], tuple):
            filename, index = newdata[0]
            filename = nexus.splitUri(filename)[0]
            file_format = os.path.splitext(filename)[-1][1:]
            dproxy.add_external(newdata, file_format=file_format)
        else:
            newdata = numpy.array(newdata)
            dproxy.add_internal(newdata)


def mem_data_generator(detector_shape):
    ndim = len(detector_shape)
    if ndim:
        x = numpy.mgrid[tuple(slice(None, n) for n in detector_shape)]
        n = numpy.prod(detector_shape)
        x = x.reshape((ndim, n))
        mu = numpy.array(detector_shape) // 2
        sigma = numpy.array(detector_shape) / 6
        cov = sigma[:, numpy.newaxis].dot(sigma[numpy.newaxis, :]) / 10
        cov[range(ndim), range(ndim)] *= 10
        data = ndgaussian(x, mu, cov).reshape(detector_shape)
        m = 1
        while True:
            yield m * data
            m += 1
    else:
        m = 1
        while True:
            yield m
            m += 1


def hdf5_data_generator(tmpdir, basename, detector_shape, nframes=3, fulluri=False):
    fmt = str(tmpdir.join(basename + "{:04d}.hdf5"))
    filename_generator = name_generator(fmt)
    data_generator = mem_data_generator(detector_shape)
    iframe = nframes
    shape = (nframes,) + detector_shape
    while True:
        data = next(data_generator)
        if iframe == nframes:
            datafile = next(filename_generator)
            with nexus.nxRoot(datafile, mode="w") as nxroot:
                nxentry = nexus.nxEntry(nxroot, "entry0000")
                nxinstrument = nexus.nxInstrument(nxentry)
                nxdetector = nexus.nxDetector(nxinstrument, "detector")
                dtype = numpy.array(data).dtype
                dset = nxdetector.create_dataset(
                    "data", shape=shape, dtype=dtype, chunks=True, fillvalue=numpy.nan
                )
                dsetname = dset.name
                nexus.markDefault(dset)
                iframe = 0
                nxroot[dsetname][iframe] = data
        else:
            with nexus.nxRoot(datafile, mode="a") as nxroot:
                nxroot[dsetname][iframe] = data
        if fulluri:
            yield datafile + "::" + dsetname, iframe
        else:
            yield datafile, iframe
        iframe += 1


def edf_data_generator(tmpdir, basename, detector_shape, nframes=3):
    fmt = str(tmpdir.join(basename + "{:04d}.edf"))
    filename_generator = name_generator(fmt)
    if not detector_shape:
        detector_shape = 1, 1
    elif len(detector_shape) == 1:
        detector_shape = 1, detector_shape[0]
    data_generator = mem_data_generator(detector_shape)
    iframe = nframes
    while True:
        data = next(data_generator)
        if iframe == nframes:
            datafile = next(filename_generator)
            edffile = EdfImage(data=data)
            iframe = 0
        else:
            edffile.append_frame(data=data)
        edffile.write(datafile)
        yield datafile, iframe
        iframe += 1


def name_generator(fmt):
    i = 0
    while True:
        yield fmt.format(i)
        i += 1


def ndgaussian(x, mu, cov, normalize=False):
    """
    :param array x: ndim x n
    :param array mu: ndim
    :param array cov: ndim x ndim
    :param bool normalize:
    :returns array: n
    """
    ndim = x.shape[0]
    if ndim != len(mu) or (ndim, ndim) != cov.shape:
        raise ValueError("The dimensions of the input don't match")
    icov = numpy.linalg.inv(cov)
    result = x - mu[:, numpy.newaxis]
    result = result * icov.dot(result)
    result = numpy.exp(-result.sum(axis=0) / 2)
    if normalize:
        result /= (2 * numpy.pi) ** (ndim / 2) * numpy.linalg.det(cov) ** 0.5
    return result


def hdf5_compression_name(dset: h5py.Dataset) -> Set[str]:
    dcpl = dset.id.get_create_plist()
    filters = set()
    for index in range(dcpl.get_nfilters()):
        info = dcpl.get_filter(index)
        filters.add(info[-1].decode())
    return filters



import numpy
import random
import itertools

from blisswriter.utils import array_order
from .helpers.nxw_test_math import asproduct


def test_order_ravel():
    for shape, order, caxes in parameter_generator():
        _test_order_ravel(shape, order, caxes)


def test_order_swap():
    for shape, order, caxes in parameter_generator():
        _test_order_swap(shape, order, caxes)


def test_order_reshape():
    for shape, order, caxes in parameter_generator():
        _test_order_reshape(shape, order, caxes)


def parameter_generator():
    for ndim in [1, 2, 3, 4]:
        for order in ["C", "F"]:
            for ncaxis in range(ndim):
                shape = tuple(random.sample(range(1, 10), ndim))
                yield shape, order, "all"
                for caxes in itertools.permutations(range(ndim), ncaxis):
                    shape = tuple(random.sample(range(1, 10), ndim))
                    yield shape, order, caxes


def _test_order_ravel(shape, order, caxes):
    err_msg = "order:{}, shape:{}, caxes:{}".format(order, shape, caxes)
    ind = list(range(numpy.prod(shape)))

    # Test some internals
    o = array_order.Order(order)
    mind1 = o.unravel(ind, shape)
    mind2 = o._unravel(ind, shape)
    numpy.testing.assert_array_equal(mind1, mind2, err_msg=err_msg)
    ind1 = o.ravel(mind1, shape)
    ind2 = o._ravel(mind2, shape)
    numpy.testing.assert_array_equal(ind, ind1, err_msg=err_msg)
    numpy.testing.assert_array_equal(ind, ind2, err_msg=err_msg)

    # Test unravel
    o = array_order.Order(order, caxes=caxes)
    midx = o.unravel(ind, shape)
    mind2 = _mindex(shape, order, caxes)
    numpy.testing.assert_array_equal(midx, mind2, err_msg=err_msg)

    # Test ravel
    ind2 = o.ravel(midx, shape)
    numpy.testing.assert_array_equal(ind2, ind, err_msg=err_msg)


def _test_order_swap(shape, order1, caxes1):
    err_msg = "order:{}, shape:{}, caxes:{}".format(order1, shape, caxes1)
    if order1 == "C":
        order2 = "F"
    else:
        order2 = "C"
    if caxes1 == "all":
        caxes2 = None
    else:
        caxes2 = tuple(i for i in range(len(shape)) if i not in caxes1)

    o1 = array_order.Order(order1, caxes=caxes1)
    o2 = array_order.Order(order2, caxes=caxes2)

    ind = list(range(numpy.prod(shape)))
    random.shuffle(ind)
    for a, b in [(o1, o2), (o2, o1)]:
        mind1 = a.unravel(ind, shape)
        mind2 = a.swap_mindex(mind1, shape, b)
        mind3 = b.unravel(ind, shape)
        numpy.testing.assert_array_equal(mind2, mind3, err_msg=err_msg)

    mind = mind1
    for a, b in [(o1, o2), (o2, o1)]:
        ind1 = a.ravel(mind, shape)
        ind2 = a.swap_index(ind1, shape, b)
        ind3 = b.ravel(mind, shape)
        numpy.testing.assert_array_equal(ind2, ind3, err_msg=err_msg)


def _test_order_reshape(shape, order, caxes):
    err_msg = "order:{}, shape:{}, caxes:{}".format(order, shape, caxes)

    o = array_order.Order(order, caxes=caxes)
    arr1 = numpy.zeros(shape)
    gen = _mindex_generator(shape, order, caxes)
    for v, idx in enumerate(gen):
        arr1[idx] = v

    # Flatten
    arr2 = numpy.arange(arr1.size)
    arr3 = o.flatten(arr1)
    numpy.testing.assert_array_equal(arr2, arr3, err_msg=err_msg)

    # Reshape
    shapes = list(asproduct(arr1.size, 2))
    if len(shapes) > 10:
        shapes = random.sample(shapes, 10)
    for newshape in shapes:
        err_msg2 = "{}, newshape: {}".format(err_msg, newshape)
        arr2 = numpy.zeros(newshape)
        gen = _mindex_generator(newshape, order, caxes)
        for v, idx in enumerate(gen):
            arr2[idx] = v
        arr3 = o.reshape(arr1, newshape)
        numpy.testing.assert_array_equal(arr2, arr3, err_msg=err_msg2)


def _mindex(shape, order, caxes):
    """Multi index in fill order (flatten gives range(0, size))"""
    gen = _mindex_generator(shape, order, caxes)
    return tuple(zip(*(idx for idx in gen)))


def _mindex_generator(shape, order, caxes):
    """Generate multi index in fill order (flatten gives range(0, size))"""
    if caxes:
        if caxes == "all":
            caxes = tuple(range(len(shape)))
        if order == "F":
            maxdim = len(shape) - 1
            flipstate = {maxdim - dim: False for dim in caxes}
        else:
            flipstate = {dim: False for dim in caxes}
    else:
        flipstate = {}
    if order == "F":
        gen = _ndindex_generator(shape[::-1], 0, flipstate)
        for idx in gen:
            yield idx[::-1]
    else:
        gen = _ndindex_generator(shape, 0, flipstate)
        for idx in gen:
            yield idx


def _ndindex_generator(shape, dim, flipstate):
    """Generate multi index of shape[dim:] in C-order"""
    if dim == len(shape) - 1:
        for i in _range(shape, dim, flipstate):
            yield (i,)
    else:
        for i in _range(shape, dim, flipstate):
            for idx in _ndindex_generator(shape, dim + 1, flipstate):
                yield (i,) + idx


def _range(shape, dim, flipstate):
    n = shape[dim]
    flip = flipstate.get(dim, None)
    if flip is not None:
        flipstate[dim] = not flip
    if flip:
        return range(n - 1, -1, -1)
    else:
        return range(n)



import os
import sys
import io
import logging
from typing import Optional
from logging.handlers import RotatingFileHandler
from inspect import signature
from ..patching import monkey


def _iter_fileio(obj, lastresort=False):
    """Yield the raw IO objects (FileIO) associated to the logger and log handlers."""
    if hasattr(obj, "handlers"):
        if obj.hasHandlers():
            for h in obj.handlers:
                yield from _iter_fileio(h)
            if obj.propagate and obj.parent is not None:
                yield from _iter_fileio(obj.parent)
        elif lastresort:
            # By default sys.stderr with WARNING level
            yield from _iter_fileio(logging.lastResort)
    elif hasattr(obj, "logger"):
        yield from _iter_fileio(obj.logger, lastresort=lastresort)
    elif hasattr(obj, "stream"):
        yield from _iter_fileio(obj.stream)
    elif hasattr(obj, "buffer"):
        yield from _iter_fileio(obj.buffer)
    elif hasattr(obj, "raw"):
        yield from _iter_fileio(obj.raw)
    elif isinstance(obj, io.FileIO):
        yield obj


class LoggingRawIO(io.RawIOBase):
    """
    Raw IO buffer that redirects writes to a logger instance.
    """

    def __init__(self, logger, log_level=logging.INFO):
        self.logger = logger
        self.log_level = log_level

    @property
    def log_level(self):
        # Make sure the output is always visible
        return max(self.logger.getEffectiveLevel(), self._log_level)

    @log_level.setter
    def log_level(self, value):
        self._log_level = value

    @property
    def name(self):
        return str(self.logger)

    def write(self, buf):
        """
        This is write(... + terminator) + flush on the logger handlers
        """
        written = len(buf)
        if isinstance(buf, memoryview):
            buf = buf.tobytes()
        self.logger.log(self.log_level, buf.decode())
        return written

    def writable(self):
        return True


def textstream_wrapper(logger, log_level=logging.INFO):
    """
    Create a stream wrapper (like sys.stdout) around a logger

    Note that if the logger writes to the stream with a terminator,
    stream.writer("abc") becomes stream.writer("abc\n"). It also flushes.
    """
    raw = LoggingRawIO(logger, log_level=log_level)
    return io.TextIOWrapper(io.BufferedWriter(raw))


# Stream that sends INFO messages to the output logger
_out_logger = logging.getLogger("OUT")
_out_logger.propagate = False
_out_logger.setLevel(logging.DEBUG)
out_stream = textstream_wrapper(_out_logger, logging.INFO)
# Stream that sends ERROR messages to the error logger
_err_logger = logging.getLogger("ERR")
_err_logger.propagate = False
_err_logger.setLevel(logging.WARNING)
err_stream = textstream_wrapper(_err_logger, logging.ERROR)


def print_out(*objects, file=None, flush=True, **kwargs):
    """
    Like builtin print but print to _out_logger by default
    """
    if file is None:
        file = out_stream
    print(*objects, file=file, flush=flush, **kwargs)


def print_err(*objects, file=None, flush=True, **kwargs):
    """
    Like builtin print but print to _err_logger by default
    """
    if file is None:
        file = err_stream
    print(*objects, file=file, flush=flush, **kwargs)


def _out_filter(level):
    """
    For logging handlers that show all but error messages

    :param int level:
    :returns: bool
    """
    return level < logging.WARNING


def _err_filter(level):
    """
    For logging handlers that only show error messages

    :param int level:
    :returns: bool
    """
    return level >= logging.WARNING


def add_cli_args(parser, default="WARNING"):
    parser.add_argument(
        "--log",
        default=default,
        type=str.upper,
        dest="level",
        help="Log level ({} by default)".format(repr(default)),
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
    )
    parser.add_argument(
        "--logfile", default="", type=str, help="Logging output to this file"
    )
    parser.add_argument(
        "--logfileout", default="", type=str, help="Logging level<WARNING to this file"
    )
    parser.add_argument(
        "--logfileerr", default="", type=str, help="Logging level>=WARNING to this file"
    )
    parser.add_argument(
        "--filesize",
        dest="maxMBytes",
        default=5,
        type=float,
        help="Maximal file size in MB",
    )
    parser.add_argument(
        "--filebackups",
        dest="backupCount",
        default=9,
        type=int,
        help="Maximal backups of each file",
    )
    parser.add_argument(
        "--nologstdout",
        action="store_false",
        dest="logstdout",
        help="No logging output to stdout",
    )
    # Not safe
    # parser.add_argument(
    #    "--nologstderr",
    #    action="store_false",
    #    dest="logstderr",
    #    help="No logging output to stderr",
    # )
    parser.add_argument(
        "--redirectstdout",
        action="store_true",
        help="Redirect stdout to a dedicated logger",
    )
    parser.add_argument(
        "--redirectstderr",
        action="store_true",
        help="Redirect stderr to a dedicated logger",
    )


def cliconfig(logger, args):
    """Configure logging from command-line options"""
    args.logstderr = True
    keys = [
        p.name
        for p in signature(config).parameters.values()
        if p.kind == p.POSITIONAL_OR_KEYWORD
    ]
    kwargs = {name: getattr(args, name) for name in vars(args) if name in keys}
    config(logger, **kwargs)


DEFAULT_FORMAT = "%(levelname)s %(asctime)s %(name)s: %(message)s"


def config(
    logger,
    level: str = None,
    logfile: Optional[str] = None,
    logfileout: Optional[str] = None,
    logfileerr: Optional[str] = None,
    logstdout: bool = True,
    logstderr: bool = True,
    redirectstdout: bool = False,
    redirectstderr: bool = False,
    maxMBytes: float = 5,
    backupCount: int = 9,
):
    # Set the logger's log level: filters messages before
    # sending it the the handlers, which themselves can
    # have a filter
    logger.setLevel(level.upper())

    # Redirect stdout and stderr to out_stream and err_stream respectively
    monkey.patch("std", stdout=redirectstdout, stderr=redirectstderr)
    org_stdout = monkey.original(sys, "stdout")
    org_stderr = monkey.original(sys, "stderr")

    # logger output handler ...
    kwargs = {
        "maxMBytes": maxMBytes,
        "backupCount": backupCount,
        "terminator": True,
        "filter_func": None,
        "formatter": True,
    }

    # - all messages to ...
    kwargs["filter_func"] = None
    if logfile:
        _add_filehandler(logger, logfile, **kwargs)

    # - all but error messages to ...
    kwargs["filter_func"] = _out_filter
    if logstdout:
        _add_streamhandler(logger, org_stdout, **kwargs)
    if logfileout:
        _add_filehandler(logger, logfileout, **kwargs)

    # - all error messages to ...
    kwargs["filter_func"] = _err_filter
    if logstderr:
        _add_streamhandler(logger, org_stderr, **kwargs)
    if logfileerr:
        _add_filehandler(logger, logfileerr, **kwargs)

    # output and error logger output handlers ...
    kwargs = {
        "maxMBytes": maxMBytes,
        "backupCount": backupCount,
        "terminator": False,
        "filter_func": None,  # already done by the loggers
        "formatter": False,
    }

    # - output and error logger to ...
    if logfile:
        _add_filehandler(_out_logger, logfile, **kwargs)
        _add_filehandler(_err_logger, logfile, **kwargs)

    # - output logger to ...
    if logstdout:
        _add_streamhandler(_out_logger, org_stdout, **kwargs)
    if logfileout:
        _add_filehandler(_out_logger, logfileout, **kwargs)

    # - error logger to ...
    if logstderr:
        _add_streamhandler(_err_logger, org_stderr, **kwargs)
    if logfileerr:
        _add_filehandler(_err_logger, logfileerr, **kwargs)


class _StreamHandlerNT(logging.StreamHandler):
    terminator = ""


class _FileHandlerNT(logging.FileHandler):
    terminator = ""


class _RotatingFileHandlerNT(RotatingFileHandler):
    terminator = ""


def _add_streamhandler(
    logger, stream, terminator=True, maxMBytes=None, backupCount=None, **kwargs
):
    """Direct logger output to a stream"""
    if logger.hasHandlers():
        add = set(_iter_fileio(stream))
        existing = set(_iter_fileio(logger))
        if not add - existing:
            return
    if terminator:
        outputhandler = logging.StreamHandler(stream)
    else:
        outputhandler = _StreamHandlerNT(stream)
    _add_outputhandler(logger, outputhandler, **kwargs)
    return outputhandler


def _add_filehandler(
    logger, filename: str, terminator=True, maxMBytes=5, backupCount=9, **kwargs
):
    """Direct logger output to a file"""
    filename = os.path.abspath(filename)
    if logger.hasHandlers():
        existing = set(stream.name for stream in _iter_fileio(logger))
        if filename in existing:
            return
    if maxMBytes:
        maxBytes = maxMBytes * 1024 * 1024
        if terminator:
            outputhandler = RotatingFileHandler(
                filename, maxBytes=maxBytes, backupCount=backupCount
            )
        else:
            outputhandler = _RotatingFileHandlerNT(
                filename, maxBytes=maxBytes, backupCount=backupCount
            )
    else:
        if terminator:
            outputhandler = logging.FileHandler(filename)
        else:
            outputhandler = _FileHandlerNT(filename)
    _add_outputhandler(logger, outputhandler, **kwargs)
    return outputhandler


def _add_outputhandler(
    logger, outputhandler, formatter=False, filter_func=None, fmt=DEFAULT_FORMAT
):
    """Direct logger output to an output handler"""
    if formatter:
        _add_formatting(outputhandler, fmt=fmt)
    if filter_func is not None:
        _add_levelfilter(outputhandler, filter_func)
    outputhandler.addFilter(_ShortLevelFilter())
    logger.addHandler(outputhandler)


def _add_formatting(outputhandler, fmt=DEFAULT_FORMAT):
    """Define format of the logger output"""
    outputhandler.setFormatter(logging.Formatter(fmt))


def _add_levelfilter(outputhandler, level_filter):
    """Filter of the logger output"""
    outputhandler.addFilter(lambda record: level_filter(record.levelno))


class _ShortLevelFilter(logging.Filter):
    NAMEMAP = {"WARNING": "WARN ", "CRITICAL": "FATAL", "INFO": "INFO "}

    def filter(self, record):
        levelname = record.levelname
        record.levelname = self.NAMEMAP.get(levelname, levelname)
        return True


class CustomLogger(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        return f"[{self.extra}] {msg}", kwargs


def _filestream(name, filename, **kwargs):
    """Useful for rollover files"""
    logger = logging.getLogger(name)
    logger.propagate = False
    logger.setLevel(logging.INFO)
    _add_filehandler(logger, filename, terminator=False, **kwargs)
    return textstream_wrapper(logger, logging.INFO)


def log(logger, msg, force=True):
    """
    :param logger: use the default stdout logger when None
    :param str msg:
    :param bool force: visible regardless of the log level
    """
    if logger is None:
        print_out(msg)
    else:
        if force:
            level = max(logger.getEffectiveLevel(), logging.INFO)
        else:
            level = logging.INFO
        logger.log(level, msg)



import queue
from blisswriter.subscribers.base_subscriber import BaseSubscriber


class _TestSubscriber(BaseSubscriber):
    def __init__(self, q):
        super().__init__("test")
        self._q = q
        self.counter = 0

    def _run(self):
        super()._run()
        while not self._stop_requested:
            try:
                item = self._q.get(timeout=0.1)
            except queue.Empty:
                continue
            if item == "end":
                break
            if item == "error":
                raise RuntimeError("test failure")
            self.counter += 1

    def __enter__(self):
        return self

    def __exit__(self, *_):
        """Ensure threads are cleaned up"""
        self.stop(timeout=3)


def test_base_subscriber():
    q = queue.Queue()
    with _TestSubscriber(q) as subscriber:
        assert subscriber.state == subscriber.state.INIT

        subscriber.start(timeout=3)
        assert subscriber.state == subscriber.state.ON
        for i in range(5):
            q.put(i)
        q.put("end")
        subscriber.join(timeout=3)

        assert subscriber.state == subscriber.state.OFF
        assert subscriber.counter == 5


def test_base_subscriber_stop():
    q = queue.Queue()
    with _TestSubscriber(q) as subscriber:
        subscriber.start(timeout=3)
        assert subscriber.state == subscriber.state.ON
        subscriber.stop(timeout=3)

        assert subscriber.state == subscriber.state.OFF


def test_base_subscriber_fault():
    q = queue.Queue()
    with _TestSubscriber(q) as subscriber:
        subscriber.start(timeout=3)
        q.put("error")
        subscriber.join(timeout=3)

        assert subscriber.state == subscriber.state.FAULT
        assert subscriber.state_reason == "test failure"


def test_base_subscriber_sort():
    q = queue.Queue()
    with _TestSubscriber(q) as subscriber1:
        with _TestSubscriber(q) as subscriber2:
            subscriber2.start(timeout=3)
        assert [subscriber1, subscriber2] == sorted([subscriber2, subscriber1])



import enum
import logging
import datetime
from collections import OrderedDict
from contextlib import contextmanager
from typing import Generator, Dict, Optional, Tuple, List, Any

from blissdata.beacon.data import BeaconData
from blissdata.redis_engine.store import DataStore
from blissdata.redis_engine.scan import Scan as RedisScan
from blissdata.redis_engine.exceptions import NoScanAvailable

from ..parameters.base import PROFILE_PARAMETERS
from .base_subscriber import BaseSubscriber
from .scan_subscriber import NexusScanSubscriber
from ..utils.periodic import PeriodicTask
from ..utils import process_utils


_logger = logging.getLogger(__name__)


class NexusSessionSubscriber(BaseSubscriber):
    @enum.unique
    class _STATE(enum.IntEnum):
        """
        * INIT: initializing (not accepting scans yet)
        * ON: accepting scans (without active scan subscribers)
        * RUNNING: accepting scans (with active scan subscribers)
        * OFF: not accepting scans
        * FAULT: not accepting scans due to exception
        """

        INIT = enum.auto()
        ON = enum.auto()
        RUNNING = enum.auto()
        FAULT = enum.auto()
        OFF = enum.auto()

    _ALLOWED_STATE_TRANSITIONS = {
        _STATE.INIT: [_STATE.INIT, _STATE.ON, _STATE.RUNNING, _STATE.OFF, _STATE.FAULT],
        _STATE.ON: [_STATE.RUNNING, _STATE.OFF, _STATE.FAULT],
        _STATE.RUNNING: [_STATE.ON, _STATE.OFF, _STATE.FAULT],
        _STATE.OFF: [_STATE.OFF, _STATE.FAULT],
        _STATE.FAULT: [],
    }

    def __init__(
        self,
        session_name: str,
        parent_logger=None,
        purge_delay: float = 300,
        resource_profiling: Optional[PROFILE_PARAMETERS] = None,
        start_semaphore_file: Optional[str] = None,
        **save_options,
    ) -> None:
        self._save_options = save_options
        self._purge_delay: float = purge_delay
        self._minimal_purge_delay: float = 5
        self._scan_subscribers: Dict[str, NexusScanSubscriber] = dict()
        self._periodic_tasks = [
            PeriodicTask(self.purge_scan_writers, 0),
            PeriodicTask(self._log_resources, 10),
        ]
        self._yield_period: float = 1

        if parent_logger is None:
            parent_logger = _logger
        super().__init__(
            session_name,
            parent_logger=parent_logger,
            resource_profiling=resource_profiling,
            start_semaphore_file=start_semaphore_file,
        )

    @property
    def _state(self):
        if self.__state == self._STATE.ON and any(
            scan_subscriber.is_alive()
            for scan_subscriber in list(self._scan_subscribers.values())
        ):
            return self._STATE.RUNNING
        return self.__state

    @_state.setter
    def _state(self, value):
        self.__state = value

    def stop(self, block: bool = True, timeout: Optional[float] = None) -> None:
        if self.state == self._STATE.RUNNING:
            raise RuntimeError("Cannot stop session subscriber when scans are running")
        super().stop(block=block, timeout=timeout)

    def _run(self):
        redis_url = BeaconData().get_redis_data_db()
        self._data_store = DataStore(redis_url)
        since = self._data_store.get_last_scan_timetag()
        super()._run()
        for scan in self._iter_scans(since):
            scan_subscriber = NexusScanSubscriber(
                scan, parent_logger=self._logger, **self._save_options
            )
            scan_subscriber.start(block=False)
            self._scan_subscribers[scan_subscriber.name] = scan_subscriber

    def _iter_scans(self, since: Optional[str]) -> Generator[RedisScan, None, None]:
        """Iterator stops only when the `stop` method is called."""
        while not self._stop_requested:
            try:
                since, key = self._data_store.get_next_scan(
                    since=since, timeout=self._yield_period
                )
                scan = self._data_store.load_scan(key, scan_cls=RedisScan)
                if scan.session == self.name:
                    yield scan
            except NoScanAvailable:
                pass
            self._execute_periodic_tasks()

    def _execute_periodic_tasks(self):
        for task in self._periodic_tasks:
            task.execute()

    def purge_scan_writers(self, delay=True):
        if delay:
            delay = max(self._minimal_purge_delay, self._purge_delay)
        else:
            delay = self._minimal_purge_delay
        scan_subscribers = dict()
        for name, scan_subscriber in list(self._scan_subscribers.items()):
            if scan_subscriber.done(delay):
                self._logger.info("Purge scan subscriber %s", scan_subscriber)
            else:
                scan_subscribers[name] = scan_subscriber
        self._scan_subscribers = scan_subscribers

    def update_saveoptions(self, **kwargs):
        self._save_options.update(kwargs)

    @property
    def resource_profiling(self) -> PROFILE_PARAMETERS:
        return self._save_options["resource_profiling"]

    @resource_profiling.setter
    def resource_profiling(self, value):
        if value is None:
            self._save_options["resource_profiling"] = PROFILE_PARAMETERS.OFF
        else:
            self._save_options["resource_profiling"] = PROFILE_PARAMETERS(value)
        self._log_resources()

    def _log_resources(self):
        if self.resource_profiling == PROFILE_PARAMETERS.OFF:
            return
        n = len(self._scan_subscribers)
        nactive = sum(w.is_alive() for w in list(self._scan_subscribers.values()))
        self._logger.info("%d scan subscribers (%d active)", n, nactive)
        self._logger.info("%s", _GLOBAL_RESOURCES)

    def get_resources(self) -> str:
        return str(_GLOBAL_RESOURCES)

    @contextmanager
    def _possible_profiling(self) -> Generator[None, None, None]:
        yield

    def _get_scan_subscriber(
        self, name: str, raise_non_existing: bool = True
    ) -> Optional[NexusScanSubscriber]:
        scan_subscriber = self._scan_subscribers.get(name, None)
        if scan_subscriber is None:
            if raise_non_existing:
                raise ValueError(f"No subscriber for scan {repr(name)} exists")
        return scan_subscriber

    def _sorted_subscriber_items(self) -> List[Tuple[str, NexusScanSubscriber]]:
        return sorted(list(self._scan_subscribers.items()), key=lambda item: item[1])

    def _scan_properties(self, getter, name=None) -> Dict[str, Any]:
        ret = OrderedDict()
        if name:
            subscriber = self._get_scan_subscriber(name)
            ret[name] = getter(subscriber)
        else:
            for name, subscriber in self._sorted_subscriber_items():
                ret[name] = getter(subscriber)
        return ret

    def scan_exists(self, name: str) -> bool:
        return name in self._scan_subscribers

    def scan_names(self) -> List[str]:
        return [name for name, _ in self._sorted_subscriber_items()]

    def stop_scan_writer(self, name: str) -> None:
        subscriber = self._get_scan_subscriber(name, raise_non_existing=False)
        if subscriber is not None:
            subscriber.stop()

    def scan_has_write_permissions(self, name: str) -> bool:
        subscriber = self._get_scan_subscriber(name)
        return subscriber.scan_writer.has_write_permissions()

    def scan_has_required_disk_space(self, name: str) -> bool:
        return not self.scan_disk_space_error(name)

    def scan_disk_space_error(self, name: str) -> str:
        subscriber = self._get_scan_subscriber(name)
        return subscriber.scan_writer.disk_space_error()

    def scan_disk_space_warning(self, name: str) -> str:
        subscriber = self._get_scan_subscriber(name)
        return subscriber.scan_writer.disk_space_warning()

    def scan_state(
        self, name: Optional[str] = None
    ) -> Dict[str, NexusScanSubscriber._STATE]:
        return self._scan_properties(lambda s: s.state, name=name)

    def scan_state_info(
        self, name: Optional[str] = None
    ) -> Dict[str, Tuple[NexusScanSubscriber._STATE, Optional[str]]]:
        return self._scan_properties(lambda s: (s.state, s.state_reason), name=name)

    def scan_urls(self, name: Optional[str] = None) -> Dict[str, List[str]]:
        return self._scan_properties(
            lambda s: s.scan_writer.scan_info.get_urls(), name=name
        )

    def scan_start(
        self, name: Optional[str] = None
    ) -> Dict[str, List[Optional[datetime.datetime]]]:
        return self._scan_properties(lambda s: s.start_time, name=name)

    def scan_end(
        self, name: Optional[str] = None
    ) -> Dict[str, List[Optional[datetime.datetime]]]:
        return self._scan_properties(lambda s: s.end_time, name=name)

    def scan_duration(
        self, name: Optional[str] = None
    ) -> Dict[str, List[datetime.timedelta]]:
        return self._scan_properties(lambda s: s.get_duration(), name=name)

    def scan_progress(self, name: Optional[str] = None) -> Dict[str, Dict[str, str]]:
        return self._scan_properties(lambda s: s.get_progress(), name=name)

    def scan_progress_info(
        self, name: Optional[str] = None
    ) -> Dict[str, Dict[str, str]]:
        return self._scan_properties(lambda s: s.get_progress_info(), name=name)


class _GlobalResources:
    def __str__(self) -> str:
        nfds = len(process_utils.file_descriptors())
        nsockets = len(process_utils.sockets())
        ngreenlets = len(process_utils.greenlets())
        nthreads = len(process_utils.threads())
        mb = int(process_utils.memory() / 1024**2)
        return f"{nthreads} threads, {ngreenlets} greenlets, {nsockets} sockets, {nfds} fds, {mb}MB MEM"


_GLOBAL_RESOURCES = _GlobalResources()



"""NeXus writer save options"""

from typing import Dict, Any
from . import base
from . import config

cli_saveoptions = {
    "noconfig": {
        "dest": "configurable",
        "action": "store_false",
        "help": "Do not use extra writer information from Redis",
    },
    "start_semaphore_file": {
        "dest": "start_semaphore_file",
        "default": None,
        "type": str,
        "help": "Create this file when started listening",
    },
}


def all_cli_saveoptions(configurable: bool = True) -> Dict[str, Dict[str, Any]]:
    if configurable:
        ret = dict(config.cli_saveoptions)
    else:
        ret = dict(base.cli_saveoptions)
    ret.update(cli_saveoptions)
    return ret


def default_saveoptions(configurable: bool = False) -> Dict[str, Any]:
    if configurable:
        ret = config.default_saveoptions()
    else:
        ret = base.default_saveoptions()
    ret["configurable"] = configurable
    return ret



def iter_chained_exception(error, include_top=True):
    if include_top:
        yield error
    parent = error.__cause__  # explicitly chained
    while parent is not None:
        yield parent
        parent = parent.__cause__
    parent = error.__context__  # implicitly chained
    while parent is not None:
        yield parent
        parent = parent.__context__



import time
import random
import multiprocessing
from concurrent.futures import ThreadPoolExecutor

import pytest

try:
    import gevent
    import gevent.monkey
except ImportError:
    gevent = None

from blisswriter.utils import sync_utils


@pytest.mark.skipif(gevent is None, reason="Requires gevent")
def test_shared_lock_pool_gevent():
    keys = list(range(4))
    lockpool = sync_utils.SharedLockPool(timeout=3)

    def worker():
        for _ in range(100):
            for key in keys:
                with lockpool.acquire(key):
                    gevent.sleep(random.uniform(0, 1e-6))

    glts = [gevent.spawn(worker) for _ in range(100)]
    try:
        gltsdone = gevent.joinall(glts, raise_error=True, timeout=10)
        assert len(glts) == len(gltsdone)
        assert len(lockpool) == 0
    finally:
        gevent.killall(glts)


def test_shared_lock_pool_threading():
    ctx = multiprocessing.get_context("spawn")
    p = ctx.Process(target=_assert_lockpool)
    p.start()
    try:
        p.join(timeout=10)
        if p.is_alive():
            raise TimeoutError("test timed outed")
        if p.exitcode:
            raise RuntimeError("test failed")
    finally:
        p.kill()


def _assert_lockpool():
    assert not gevent.monkey.is_anything_patched()

    keys = list(range(4))
    lockpool = sync_utils.SharedLockPool(timeout=3)

    def worker(*_):
        for _ in range(100):
            for key in keys:
                with lockpool.acquire(key):
                    with lockpool.acquire(key):
                        time.sleep(random.uniform(0, 1e-6))

    with ThreadPoolExecutor(50) as pool:
        list(pool.map(worker, range(50)))

    assert len(lockpool) == 0



import logging
from tango import LogLevel, Logging

# For the TANGO loggers

tango_log_level = {
    logging.DEBUG: 600,
    logging.INFO: 500,
    logging.WARNING: 400,
    logging.ERROR: 300,
    logging.CRITICAL: 200,
    logging.NOTSET: 100,
}

itango_log_level = {v: k for k, v in tango_log_level.items()}

# For the TANGO command-line-interface

tango_cli_slog_level = {
    "DEBUG": int(LogLevel.LOG_DEBUG),
    "INFO": int(LogLevel.LOG_INFO),
    "WARNING": int(LogLevel.LOG_WARN),
    "ERROR": int(LogLevel.LOG_ERROR),
    "CRITICAL": int(LogLevel.LOG_FATAL),
    "NOTSET": int(LogLevel.LOG_OFF),
}

tango_cli_log_level = {
    logging.DEBUG: int(LogLevel.LOG_DEBUG),
    logging.INFO: int(LogLevel.LOG_INFO),
    logging.WARNING: int(LogLevel.LOG_WARN),
    logging.ERROR: int(LogLevel.LOG_ERROR),
    logging.CRITICAL: int(LogLevel.LOG_FATAL),
    logging.NOTSET: int(LogLevel.LOG_OFF),
}

itango_cli_log_level = {v: k for k, v in tango_cli_log_level.items()}

# For Beacon

beacon_log_level = {
    logging.DEBUG: "DEBUG",
    logging.INFO: "INFO",
    logging.WARNING: "WARN",
    logging.ERROR: "ERROR",
    logging.CRITICAL: "ERROR",
    logging.NOTSET: "WARN",
}

ibeacon_log_level = {v: k for k, v in beacon_log_level.items()}

# For Python

log_level_name = {
    logging.DEBUG: "DEBUG",
    logging.INFO: "INFO",
    logging.WARNING: "WARNING",
    logging.ERROR: "ERROR",
    logging.CRITICAL: "CRITICAL",
    logging.NOTSET: "NOTSET",
}

ilog_level_name = {v: k for k, v in log_level_name.items()}

# Helper functions


def add_tango_cli_args(parser, default="WARNING"):
    parser.add_argument(
        "--log_tango",
        type=str.upper,
        default=default,
        help="Tango log level",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
    )


def init_tango_log_level(device=None):
    """
    :param Device device:
    """
    # cli  name      core      device
    # -v1  CRITICAL  WARNING   INFO
    # -v2  ERROR     WARNING   INFO
    # -v3  WARNING   WARNING   DEBUG
    # -v4  INFO      WARNING   DEBUG
    # -v5  DEBUG     DEBUG     DEBUG
    #
    # core: DebugIt, Logit, ...
    # device: debug_stream, info_stream, ...

    import argparse

    parser = argparse.ArgumentParser(add_help=False)
    add_tango_cli_args(parser)
    args, _ = parser.parse_known_args()
    level = ilog_level_name[args.log_tango]
    tango_set_log_level(level, device=device)


def tango_get_log_level(device=None):
    """
    :param Device device:
    :returns int: logging.DEBUG, logging.INFO, ...
    """
    if device:
        level = device.get_logger().get_level()
    else:
        level = Logging.get_core_logger().get_level()
    return itango_log_level[level]


def tango_set_log_level(level, device=None):
    """
    :param int level: logging.DEBUG, logging.INFO, ...
    :param Device device:
    """
    level = tango_log_level[level]
    if device:
        device.get_logger().set_level(level)
    core_logger = Logging.get_core_logger()
    if core_logger is not None:
        core_logger.set_level(level)



import sys
import builtins
import logging
from functools import wraps
from . import monkey
from ..utils import logging_utils


def _print_out(old_print, logger=None):
    if logger is None:
        stream = logging_utils.out_stream
    else:
        stream = logging_utils.textstream_wrapper(logger, logging.INFO)

    @wraps(old_print)
    def print_out(*args, file=None, **kwargs):
        if file is None:
            file = stream
        return old_print(*args, file=file, **kwargs)

    return print_out


def patch_stdout(logger=None):
    monkey.patch_item(sys, "stdout", logging_utils.out_stream)


def patch_stderr(logger=None):
    monkey.patch_item(sys, "stderr", logging_utils.err_stream)


def patch_print():
    newitem = _print_out(monkey.original(builtins, "print"))
    monkey.patch_item(builtins, "print", newitem)


def patch(logger=None, stdout=True, stderr=True):
    if stdout:
        patch_stdout()
        # When patching stdout there is no need to patch print
        # patch_print(logger=logger)
    if stderr:
        patch_stderr()


def unpatch():
    monkey.unpatch_item(sys, "stdout")
    monkey.unpatch_item(sys, "stderr")



import os
from typing import Dict, Optional
from . import session_utils
from . import data_policy


@session_utils.with_scan_saving
def session_filename(scan_saving=None) -> Optional[str]:
    """Name of the file that contains the scan data of the current BLISS session"""
    return session_utils.scan_saving_get(
        "filename", default=None, scan_saving=scan_saving
    )


@session_utils.with_scan_saving
def session_master_filenames(scan_saving=None, config: bool = True) -> Dict[str, str]:
    """Names of the files that contain links to the scan data of the current BLISS session"""
    if not config:
        return dict()
    writer_object = scan_saving.writer_object
    if not writer_object.saving_enabled():
        return dict()
    eval_dict = dict()
    root_path = scan_saving.get_cached_property("root_path", eval_dict=eval_dict)
    if writer_object.separate_scan_files:
        dataset_template = scan_saving.data_filename
    else:
        dataset_template = None
    relative_templates = data_policy.masterfile_templates(
        dataset_template=dataset_template, data_policy=scan_saving.data_policy
    )
    return {
        name: scan_saving.eval_template(
            os.path.abspath(os.path.join(root_path, s)), eval_dict=eval_dict
        )
        for name, s in relative_templates.items()
    }


@session_utils.with_scan_saving
def session_filenames(scan_saving=None, config=True):
    """
    Names of the files that contain links to the scan data (raw or as links) of the current BLISS session

    :param bliss.scanning.scan.ScanSaving scan_saving:
    :pram bool config: writer parses the extra "nexuswriter" info
    :returns list(str):
    """
    filenames = dict()
    filename = session_filename(scan_saving=scan_saving)
    if config:
        filenames.update(
            session_master_filenames(scan_saving=scan_saving, config=config)
        )
    if filename:
        filenames["dataset"] = filename
    return filenames



import numpy
import itertools
import logging
from .array_order import Order


logger = logging.getLogger(__name__)


def createSlice(start, end, stride):
    """
    :param int start: start index
    :param int end: end index
    :param stride:
    :returns slice or int: integer when stride is zero
    """
    if stride:
        end += stride
        if stride < 0 and end < 0:
            end = None
        return slice(start, end, stride)
    else:
        return start


def intListToSliceIndexing(coord, slicetwo=False, allow_negative_stride=False):
    """
    Convert lists of integers to slices and/or integers

    :param tuple(array(int)) coord:
    :yields tuple(slice or int): same length as `coord`
    """
    coord = numpy.asarray(coord)
    ndim, ncoord = coord.shape
    # print('\n')
    # print(numpy.concatenate([coord, numpy.arange(ncoord)[None,:]], axis=0))

    # Determine slices
    strides = numpy.diff(coord, axis=1)
    # print(numpy.concatenate([numpy.zeros((ndim,1),dtype=int), strides], axis=1))
    diff = numpy.diff(strides, axis=1)
    newslices = numpy.any(diff, axis=0)
    if not allow_negative_stride:
        negative_strides = numpy.any(strides < 0, axis=0)
        newslices |= negative_strides[1:]

    def getstrides(start):
        try:
            return strides[:, start]
        except IndexError:
            return numpy.ones_like(coord[:, start])

    # Create a slice from start and stop index
    def makeslice(start, end):
        _start = coord[:, start]
        _end = coord[:, end]
        _strides = getstrides(start)
        if not allow_negative_stride:
            if (_strides < 0).any():
                lst = [
                    list(range(start, end + stride, stride)) if stride else [start]
                    for start, end, stride in zip(_start, _end, _strides)
                ]
                n = max(map(len, lst))
                lst = [e * n if len(e) == 1 else e for e in lst]
                for idx in zip(*lst):
                    yield tuple(idx)
                return
        yield tuple(createSlice(*args) for args in zip(_start, _end, _strides))

    def split2(start, end):
        if slicetwo:
            return False
        if start + 1 == end:
            if allow_negative_stride:
                return True
            else:
                return (getstrides(start) < 0).any()
        else:
            return False

    # Yield the slices
    start = 0
    skip = False
    for end, newslice in enumerate(newslices, 1):
        if skip:
            skip = False
        elif newslice:
            if start + 1 == end and not slicetwo:
                yield tuple(coord[:, start])
                start = end
            else:
                for idx in makeslice(start, end):
                    yield idx
                start = end + 1
                skip = True

    # Yield the last slice
    end = ncoord - 1
    if start == end:
        yield tuple(coord[:, start])
    elif start + 1 == end and not slicetwo:
        a = tuple(coord[:, start])
        b = tuple(coord[:, end])
        yield a
        if a != b:
            yield b
    else:
        for idx in makeslice(start, end):
            yield idx


def intListToSliceIndexing2(coordrange, coord):
    """
    Convert lists of integers to slices and/or integers

    :param array(int) coordrange: a range with step one
    :param tuple(array(int)) coord:
    :yields tuple, tuple:
    """
    start = 0
    for idx in intListToSliceIndexing(coord, slicetwo=True):
        for s in idx:
            if isinstance(s, slice):
                stop = start + (s.stop - s.start) // s.step
                idxflat = slice(coordrange[start], coordrange[stop - 1] + 1, 1)
                start = stop
                break
        else:
            idxflat = coordrange[start]
            start += 1
        yield (idxflat,), idx


def _mergedShape(shapes, axis=0, newaxis=True):
    """
    Shape after merging

    :param list(tuple) shapes:
    :param int axis: merge along this axis
    :param bool newaxis: merge axis is new or existing
    :returns tuple or None:
    """
    if not shapes:
        return None
    shape = list(shapes[0])
    if newaxis:
        if axis < 0:
            axis += len(shape) + 1
        shape.insert(axis, len(shapes))
    else:
        shape[axis] = 0
        for s in shapes:
            shape[axis] += s[axis]
    return tuple(shape)


def mergeShapeGenerator(sources, shapes, mshape, axis=0, newaxis=True):
    """
    Merge I/O index generator.

    :param list(any) sources:
    :param list(tuple) shapes: of sources
    :param tuple mshape: merged shape
    :param int axis: merge along this axis
    :param bool newaxis: merge axis is new or existing
    :returns generator: index generator
    """
    logger.debug(
        "{} --{}--> {}".format(shapes, "stack" if newaxis else "concat", mshape)
    )
    if len(sources) == 1:

        def fill_generator():
            def index_generator():
                yield tuple(), tuple()

            yield sources[0], index_generator

    else:

        def fill_generator():
            idx = [slice(None)] * len(mshape)
            n = 0
            for i, (source, shapei) in enumerate(zip(sources, shapes)):
                if newaxis:
                    idx[axis] = i
                else:
                    idx[axis] = slice(n, n + shapei[axis])
                    n += shapei[axis]

                def index_generator():
                    yield tuple(), tuple(idx)

                yield source, index_generator

    return fill_generator


def _countCommonFastAxes(shapein, shapeout, newshapeout, order=None):
    """
    Full index along common fast axes.

    :param tuple shapein: slice of `shapeout`
    :param tuple shapeout:
    :param tuple newshapeout: reshape `shapeout`
    :param Order order:
    :returns tuple(int):
    """
    # Fast axis first
    if order.order != "F":
        shapein = shapein[::-1]
        shapeout = shapeout[::-1]
        newshapeout = newshapeout[::-1]

    shapein = numpy.array(shapein)
    arrin = numpy.cumprod(shapein)[:-1]
    shapeout = numpy.array(shapeout)
    arrout = numpy.cumprod(shapeout)[:-1]
    newshapeout = numpy.array(newshapeout)
    arroutnew = numpy.cumprod(newshapeout)[:-1]
    n = min([arrin.size, arrout.size, arroutnew.size])
    if not n:
        return 0, 0, 0
    b = numpy.where(
        (arrin[:n] == arroutnew[:n])
        & (arrin[:n] == arrout[:n])
        & (arrout[:n] == arroutnew[:n])
    )[0]
    if not b.size:
        return 0, 0, 0

    # Number of fast axis to skip
    m = arrin[b[-1]]
    nin = numpy.where(arrin == m)[0][-1] + 1
    nout = numpy.where(arrout == m)[0][-1] + 1
    noutnew = numpy.where(arroutnew == m)[0][-1] + 1
    return nin, nout, noutnew


def _getCoordinates(
    shape, ran, nskipfast, newshape, newnskipfast, nsources=1, order=None
):
    """
    Coordinates of all elements in `newshape` except the ones on
    the skipped fast axes.

    :param tuple shape:
    :param list(list) ran: same length as `shape`
    :param int nskipfast: number of fast axes in `shape` to skip
    :param tuple newshape: reshaped `shape`
    :param int nsources:
    :param Order order:
    :returns tuple(list): same length as `newshape`
    """
    # C: fast axis last
    # F: fast axis first
    # itertools.product: fast axis is last
    Forder = order.order == "F"
    if nskipfast:
        if Forder:
            ran = ran[nskipfast:]
            shape = shape[nskipfast:]
        else:
            ran = ran[:-nskipfast]
            shape = shape[:-nskipfast]
    if newnskipfast:
        if Forder:
            newshape = newshape[newnskipfast:]
        else:
            newshape = newshape[:-newnskipfast]
    if nsources > 1:
        if Forder:
            idx = tuple(zip(*itertools.product(*ran[::-1])))[::-1]
        else:
            idx = tuple(zip(*itertools.product(*ran)))
        flatidx = order.ravel(idx, shape)
    else:
        flatidx = numpy.arange(numpy.prod(shape))
    return order.unravel(flatidx, newshape)


def mergeReshapeGenerator(
    sources,
    shapes,
    shapeout,
    newshapeout,
    order=None,
    axis=0,
    newaxis=True,
    allow_advanced_indexing=True,
):
    """
    Reshaped merge I/O index generator.

    :param list(any) sources:
    :param list(tuple) shapes: of sources
    :param tuple shapeout: merged `shapes`
    :param tuple newshapeout: reshaping shapeout
    :param Order order: for reshaping
    :param int axis: merge along this axis
    :param bool newaxis: merge axis is new or existing
    :param bool allow_advanced_indexing:
    :returns generator: index generator
    """
    if not isinstance(order, Order):
        order = Order(order)
    logger.debug(
        "{} --{}--> {} --{}--> {}".format(
            shapes, "stack" if newaxis else "concat", shapeout, order, newshapeout
        )
    )
    nsources = len(sources)

    def fill_generator():
        noutconcat = 0
        for i, (source, shapein) in enumerate(zip(sources, shapes)):
            # Indices per dimension
            ranin = [list(range(s)) for s in shapein]
            ranout = [list(range(s)) for s in shapeout]
            if newaxis:
                ranout[axis] = [i]
            else:
                n = noutconcat + shapein[axis]
                ranout[axis] = list(range(noutconcat, n))
                noutconcat = n

            # Indices for full dimensions
            if newaxis and nsources in shapein:
                nskipin, nskipout, nskipoutnew = 0, 0, 0
            else:
                nskipin, nskipout, nskipoutnew = _countCommonFastAxes(
                    shapein, shapeout, newshapeout, order=order
                )
            fastidxin = (slice(None),) * nskipin
            fastidxoutnew = (slice(None),) * nskipoutnew

            # Indices for other dimensions
            coordin = _getCoordinates(
                shapein, ranin, nskipin, shapein, nskipin, nsources=1, order=order
            )
            coordout = _getCoordinates(
                shapeout,
                ranout,
                nskipout,
                newshapeout,
                nskipoutnew,
                nsources=nsources,
                order=order,
            )
            flatin = len(coordin) == 1
            flatout = len(coordout) == 1

            # Add full indices
            if order.order == "F":

                def fout(tpl):
                    return fastidxoutnew + tpl

                def fin(tpl):
                    return fastidxin + tpl

            else:

                def fout(tpl):
                    return tpl + fastidxoutnew

                def fin(tpl):
                    return tpl + fastidxin

            # Create index generator
            noreduction = False
            if allow_advanced_indexing:
                # Index type: list of integers
                def index_generator():
                    yield fin(coordin), fout(coordout)

            elif noreduction:
                # Index type: integers
                def index_generator():
                    for idxin, idxout in zip(zip(*coordin), zip(*coordout)):
                        yield fin(idxin), fout(idxout)

            elif flatin:

                def index_generator():
                    for idxin, idxout in intListToSliceIndexing2(coordin[0], coordout):
                        yield fin(idxin), fout(idxout)

            elif flatout:

                def index_generator():
                    for idxout, idxin in intListToSliceIndexing2(coordout[0], coordin):
                        yield fin(idxin), fout(idxout)

            else:
                # Index type: integers and slices
                n = len(coordin)
                coord = coordin + coordout

                def index_generator():
                    for idx in intListToSliceIndexing(coord):
                        yield fin(idx[:n]), fout(idx[n:])

            yield source, index_generator

    return fill_generator


def mergeGenerator(
    sources,
    shapes,
    axis=0,
    newaxis=True,
    shape=None,
    order=None,
    allow_advanced_indexing=True,
):
    """
    Equivalent to `numpy.stack` or `numpy.concatenate` combined
    with `numpy.reshape`. Return I/O index generator for merging.

    :param list(any) sources:
    :param list(tuple) shapes: of sources
    :param int axis: merge along this axis
    :param bool newaxis: merge axis is new or existing
    :param tuple shape: for reshaping
    :param Order order: for reshaping
    :param bool allow_advanced_indexing:
    :returns tuple, generator: merged shape and index generator
    """
    mshape = _mergedShape(shapes, axis=axis, newaxis=newaxis)
    if not mshape:

        def fill_generator():
            return
            yield None

    elif shape is None or shape == mshape:
        fill_generator = mergeShapeGenerator(
            sources, shapes, mshape, axis=axis, newaxis=newaxis
        )
        shape = mshape
    else:
        fill_generator = mergeReshapeGenerator(
            sources,
            shapes,
            mshape,
            shape,
            order=order,
            axis=axis,
            newaxis=newaxis,
            allow_advanced_indexing=allow_advanced_indexing,
        )
    return shape, fill_generator



import builtins
import inspect
from functools import wraps
from . import monkey


# Maps old class to new class
_CLASSES = {}


def _isinstance(old_isinstance, logger=None):
    @wraps(old_isinstance)
    def new_isinstance(obj, classes):
        if old_isinstance(obj, classes):
            return True
        if not old_isinstance(classes, tuple):
            classes = (classes,)
        # Replace registered classes and try again
        classes = tuple(
            o for _class in classes for o in _CLASSES.get(_class, (_class,))
        )
        return old_isinstance(obj, classes)

    return new_isinstance


def _get_base_classes(orgcls, newcls):
    new_classes = inspect.getmro(newcls)
    for _orgcls in inspect.getmro(orgcls):
        if _orgcls not in new_classes:
            yield _orgcls


def register_class(orgcls, newcls):
    for _orgcls in _get_base_classes(orgcls, newcls):
        s = _CLASSES.setdefault(_orgcls, set())
        s.add(newcls)
    print(_CLASSES)


def unregister_class(orgcls, newcls):
    for _orgcls in _get_base_classes(orgcls, newcls):
        s = _CLASSES.setdefault(_orgcls, set())
        s.remove(newcls)


def patch():
    newitem = _isinstance(monkey.original(builtins, "isinstance"))
    monkey.patch_item(builtins, "isinstance", newitem)


def unpatch():
    monkey.unpatch_item(builtins, "isinstance")
